#!/usr/bin/perl -w

use strict;
use FindBin;

sub BEGIN { unshift @INC, $FindBin::Bin; }

use Common;

my @args = @ARGV;

my $pkg = undef; # filename or name of installed package
my $is_installed = 0;
my $manager = undef; # 'RPM' or 'deb'
my $dir = undef;

while ( @args ) {
	my $arg = shift @args;
	
	if ( $arg eq '--pkg' ) {
		$is_installed = 1;
	}
	elsif ( $arg eq '-u' ) {
		$manager = shift @args;
		if ( $manager ne 'RPM' && $manager ne 'deb' ) {
			die usage("Package format is expected to be 'RPM' or 'deb'");
		}
	}
	elsif ( $arg eq '-d' ) {
		$dir = shift @args;
	}
	else {
		if ( $arg =~ /^-/ ) {
			die usage("Unexpected arguement: '$arg'");
		}
		$pkg = $arg;
	}
}

defined $pkg or die usage("Package name is expected.");
defined $manager or die usage("Package manager type is expected. Use -u switch.");
defined $dir or die usage("Unpack directory is expected. Use -d switch.");

if ( !-d $dir ) {
	mkdir($dir);
}
-d $dir or die "Failed to create directory '$dir'. $!";

if ( $manager eq 'RPM' ) {
	rpm_control_extract();
}
else {
	deb_control_extract();
}

exit 0;

#-----------------------------------------------------------------------

sub usage {
	my ($msg) = @_;
	
	if ( $msg ) {
		print $msg, "ERROR: \n\n";
	}
	
	print <<HEREDOC;
Extracts control scripts (preinstall, postinstall, etc.) from RPM and deb
packages.

Usage: $0 -u RPM -d ./SCRIPTS [--pkg] package

  -u		Package type: 'RPM' or 'deb'.
  -d		Directory where to unpack the pre/post-install scripts.
  			The directory (but not its parents) will be created if necessary.
  --pkg		Work with installed package.
HEREDOC
}

sub rpm_control_extract {

	my $cmd = 'rpm -q --scripts --triggers '.($is_installed ? '' : '-p ').esc_path($pkg);

	my $out = `$cmd`;
	die "Error: $!" if $? != 0;

	my $file;

	foreach my $line ( split /\n/, $out ) {
		#print $line,"\n";
		if ( $line =~ m/^(\w+) scriptlet \(using ([^\)]+)\)/ ) {
			my $script = $1;
			my $interp = $2;
			
			close $file if $file;
			open $file, ">$dir/$script" or die "Failed to open for writing: '$dir/$script', $!";
			
			print {$file} "#!$interp\n\n";
			
			next;
		}
		if ( !$file ) {
			die "Unexpected output: '$line'";
		}
		print {$file} "$line\n";
	}

	close $file if $file;
}


sub deb_control_extract {
	
	if ( $is_installed ) {
		# All the scripts are in /var/lib/dpkg/info
		
		my $glob = "/var/lib/dpkg/info/$pkg.*";
		
		for my $file ( glob $glob ) {
			my ($end) = $file =~ /\.([^\.]+)$/;
			$end ||= $file;
			my $cmd = esc_cmdline("cp «$file» «$dir$end»");
			system($cmd) == 0 or die "Failed '$cmd': $!";
		}
	} else {
		# Use dpkg-deb -e <package> <dir> command
		my $cmd = esc_cmdline("dpkg-deb -e «$pkg» «$dir»");
		system($cmd) == 0 or die "Failed '$cmd': $!";
	}
}