#!/usr/bin/perl -w

# Application Checker
# Journal Parser and Report Generator Module (report.pl)
#
# Copyright (C) 2007-2010 The Linux Foundation. All rights reserved.
#
# This program has been developed by ISP RAS for LF.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
# # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

use strict;
use Fcntl ':seek';

sub BEGIN {
	my $utils_path = $0;
	$utils_path =~ s!/[^/]+$!!;
	if ($utils_path !~ m!^/!) {
		my $cwd = `pwd`;
		$cwd =~ s/[\r\n]//g;
		$utils_path = "$cwd/$utils_path";
	}
	push @INC, $utils_path;
}

use Common;
use SimpleJournal;
use ManualTests;

autoflush_on();

################################################################################
# Global variables

# Paths
my $result_path;           # Directory containing all test results.
my $result_dir;            # Directory with the current test result.
my $result_archive;        # TAR.GZ archive name to pack the results into.
my $application_name;      # Name of the application tested.

# Host information
my $host_name;
my $host_os;
my $host_machine;
my $test_arch;
my $host_kernel;

# Test information
my $spec_version;
my $test_comments;

# List of all journal titles
my @journal_titles = ();

# List of HTML journal anchors in the same order as they appear in the tree
my @journal_anchors_sorted = ();

# Information for each journal
# %journal_data: <journal-title> => {
#                    'JOURNAL_FILE' => <file-name>
#                    'TEST_TOOL' => <tool-name>
#                    'START_TIME' => <time>
#                    'IS_RESULT' => (1|0|undef),            # undef - no journal, 0 - journal empty
#                    'PROBLEMS_LIST' => [
#                        {
#                            'SEVERITY' => (Warning|Error|Notice|...),
#                            'TEST_CASE' => <test-case ID>,
#                            'SUBCATEGORY' => <SUBCAT-ID>,
#                            'ELEMENT' => <element>,
#                            'COMMENT' => <comment>,
#                            'DETAILS' => <description-text>
#                        },
#                        ...
#                    ],
#                    'STATISTICS' => {
#                        'PASSES' => <number>,
#                        'FAILS' => <number>,
#                        'WARNS' => <number>,
#                        'FIPS' => <number>,
#                        'NOTICES' => <number>,
#                        'QUELLED' => <number>,
#                        'SKIPPED' => <number>
#                    }
#                };
my %journal_data;

# Information from manual tests journal.
my $manual_tests_data;

# Information about categories.
# %categories: CAT_ID => {
#                  'DESCRIPTION' => <text>,          # Short descriptive information shown below category title
#                  'ELEM_COL_NAME' => <name>         # Overridden title of the 'Element' column for the category table
#              };
my %categories = (
	# Problems caused by test suite itself or some environment
	'Journal format error' => {
		'DESCRIPTION' => 'Journal file seems to be corrupted.',
		'ELEM_COL_NAME' => 'Element'
	},
	'Journal syntax error' => {
		'DESCRIPTION' => 'A line in the journal file has incorrect syntax.',
		'ELEM_COL_NAME' => 'Element'
	},
	'Test case controller error' => {
		'DESCRIPTION' => 'Error reported by Test Case Controller.',
		'ELEM_COL_NAME' => 'Element'
	},
	'Test suite error' => {
		'DESCRIPTION' => 'Error reported by Test Suite itself.',
		'ELEM_COL_NAME' => 'Element'
	},
	'Empty test case' => {
		'DESCRIPTION' => 'Empty test case encountered (no test points).',
		'ELEM_COL_NAME' => 'Element'
	},
	'Test suite run aborted' => {
		'DESCRIPTION' => 'Test suite was aborted during execution by the user.',
		'ELEM_COL_NAME' => 'Element'
	},
	'Test point restarted' => {
		'DESCRIPTION' => 'Test Point Start entry appeared again for the same test point.',
		'ELEM_COL_NAME' => 'Element'
	},
	'Unrecognized result code' => {
		'DESCRIPTION' => 'Test returned a result code not recognized by the '.$appchecker_branch{'OS'}.' AppChecker.',
		'ELEM_COL_NAME' => 'Element'
	},
	'Wrong archive format' => {
		'DESCRIPTION' => 'There are problems in the archive format.',
		'ELEM_COL_NAME' => 'Element'
	},
	# Problems found in the files tested
	# appchk-elf
	'ELF header check failed' => {
		'DESCRIPTION' => 'Some ELF header fields differ from what '.$appchecker_branch{'STD'}.' declares.',
		'ELEM_COL_NAME' => 'Element'
	},
	'Dynamic segment check failed' => {
		'DESCRIPTION' => 'Problems encountered while checking PT_DYNAMIC segment (dynamic linking information).',
		'ELEM_COL_NAME' => 'Element'
	},
	'Incorrect program loader' => {
		'DESCRIPTION' => 'The application uses program loader different from what '.$appchecker_branch{'STD'}.' declares. Try compiling your application with <b>lsbcc</b> compiler from LSB SDK package.',
		'ELEM_COL_NAME' => 'Loader'
	},
	'NOTE section check failed' => {
		'DESCRIPTION' => 'Problems encountered while checking PT_NOTE segment (auxiliary information).',
		'ELEM_COL_NAME' => 'Element'
	},
	'TLS section check failed' => {
		'DESCRIPTION' => 'Problems encountered while checking PT_TLS segment (thread-local storage).',
		'ELEM_COL_NAME' => 'Element'
	},
	'GNU_STACK header check failed' => {
		'DESCRIPTION' => 'Problems encountered while checking PT_GNU_STACK segment.',
		'ELEM_COL_NAME' => 'Element'
	},
	'ELF section check failed' => {
		'DESCRIPTION' => 'Problems encountered while checking ELF sections.',
		'ELEM_COL_NAME' => 'Element'
	},
	'Non-'.$appchecker_branch{'STD'}.' ELF section' => {
		'DESCRIPTION' => 'ELF sections found that are not present in '.$appchecker_branch{'STD'}.'.',
		'ELEM_COL_NAME' => 'Element'
	},
	'Non-'.$appchecker_branch{'STD'}.' library used' => {
		'DESCRIPTION' => 'The application uses a library that is not present in '.$appchecker_branch{'STD'}.'.',
		'ELEM_COL_NAME' => 'Library'
	},
	'Non-'.$appchecker_branch{'STD'}.' interface used' => {
		'DESCRIPTION' => 'The application uses an interface that is not present in '.$appchecker_branch{'STD'}.'.',
		'ELEM_COL_NAME' => 'Interface'
	},
	'I/O problem' => {
		'DESCRIPTION' => 'Problems encountered while trying to read file contents.',
		'ELEM_COL_NAME' => 'Element'
	},
	'Unexpected interface version' => {
		'DESCRIPTION' => 'A versioned symbol is used, but the version is not '.$appchecker_branch{'STD'}.'-compliant.',
		'ELEM_COL_NAME' => 'Interface'
	},
	'Deprecated interface used' => {
		'DESCRIPTION' => 'Deprecated interface are not recommended to use, because they will soon be withdrawn from '.$appchecker_branch{'STD'}.'.',
		'ELEM_COL_NAME' => 'Interface'
	},
	'ioctl() encountered' => {
		'DESCRIPTION' => '<b>ioctl()</b> interface is used to control stream device directly. Static application analysis is unable to check whether this interface is used according to '.$appchecker_branch{'STD'}.' or not, so you need to check it yourself and provide explanatory comments.',
		'ELEM_COL_NAME' => 'Interface'
	},
	'dlopen() encountered' => {
		'DESCRIPTION' => '<b>dlopen()</b> interface is used to dynamically load shared objects at run-time. Static application analysis is unable to check whether this interface is used according to '.$appchecker_branch{'STD'}.' or not, so you need to check it yourself and provide explanatory comments.',
		'ELEM_COL_NAME' => 'Interface'
	},
	# appchk-shell (Linux AppChecker only)
	'Wrong shell script interpreter' => {
		'DESCRIPTION' => 'Shell script interpreter should be \'sh\' and located in /bin directory.',
		'ELEM_COL_NAME' => 'Interpreter'
	},
	'Bashism used in shell script' => {
		'DESCRIPTION' => $appchecker_branch{'STD'}.' declares \'sh\' the main shell program, so bashisms (features specific for \'bash\') are not allowed.',
		'ELEM_COL_NAME' => 'Element'
	},
	'Non-'.$appchecker_branch{'STD'}.' command used' => {
		'DESCRIPTION' => 'The command is not present in '.$appchecker_branch{'STD'}.'.',
		'ELEM_COL_NAME' => 'Command'
	},
	'Wrong function definition' => {
		'DESCRIPTION' => 'Function definition does not require the keyword \'function\'; bash allows it, but in sh this is error.',
		'ELEM_COL_NAME' => 'Element'
	},
	'Wrong identifier' => {
		'DESCRIPTION' => 'Invalid identifier encountered. Please use only letters, digits and underscores.',
		'ELEM_COL_NAME' => 'Identifier'
	},
	'Wrong tilde expansion' => {
		'DESCRIPTION' => 'Tilde is expanded into home directory path, but can be used in two forms only: \'~\' and \'~login\'.',
		'ELEM_COL_NAME' => 'Element'
	},
	'Ambiguous syntax' => {
		'DESCRIPTION' => 'The expression is syntactically correct but can be treated differently in different interpreters.',
		'ELEM_COL_NAME' => 'Element'
	},
	'Include error' => {
		'DESCRIPTION' => 'The script includes an external file, which isn\'t specified by the standard, and therefore it may be missing in other Linux distributions.',
		'ELEM_COL_NAME' => 'File included'
	},
	'Unconditional exit' => {
		'DESCRIPTION' => 'An unconditional exit was faced, so the rest of the file wasn\'t checked.',
		'ELEM_COL_NAME' => 'Command'
	},
	'Parse error' => {
		'DESCRIPTION' => 'appchk failed to parse the shell script.',
		'ELEM_COL_NAME' => 'Element'
	},
	# appchk-perl
	'Non-'.$appchecker_branch{'STD'}.' Perl version used' => {
		'DESCRIPTION' => 'The perl script uses Perl version not allowed by '.$appchecker_branch{'STD'}.'.',
		'ELEM_COL_NAME' => 'Version'
	},
	'Non-'.$appchecker_branch{'STD'}.' Perl module used' => {
		'DESCRIPTION' => 'The perl script uses a module not included in '.$appchecker_branch{'STD'}.'.',
		'ELEM_COL_NAME' => 'Module'
	},
	# appchk-python
	'Non-'.$appchecker_branch{'STD'}.' Python module path used' => {
		'DESCRIPTION' => 'The python script loads a module from non-standard path.',
		'ELEM_COL_NAME' => 'Module'
	},
	'Non-'.$appchecker_branch{'STD'}.' Python module used' => {
		'DESCRIPTION' => 'The python script uses a module not included in '.$appchecker_branch{'STD'}.'.',
		'ELEM_COL_NAME' => 'Module'
	},
	# desktopchk (Moblin AppChecker only)
	'Non-'.$appchecker_branch{'STD'}.' desktop path' => {
		'DESCRIPTION' => 'The application uses a .desktop file, but the installed path is not '.$appchecker_branch{'STD'}.' compliant.',
		'ELEM_COL_NAME' => 'Desktop'
	},
	'Desktop field check failed' => {
		'DESCRIPTION' => 'The application uses a .desktop file, but the field is not moblin compliant.',
		'ELEM_COL_NAME' => 'Desktop'
	},
	'Desktop icon check failed' => {
		'DESCRIPTION' => 'The application uses a .desktop file, but the icon is not moblin compliant.',
		'ELEM_COL_NAME' => 'Desktop'
	},
	# pkgchk (Moblin AppChecker only)
	'Path check failed' => {
		'DESCRIPTION' => 'File placement is not compliant to '.$appchecker_branch{'STD'}.' Specification.',
		'ELEM_COL_NAME' => 'File'
	},
	# pkgchk (Linux AppChecker only)
	'FHS check failed' => {
		'DESCRIPTION' => 'File placement is not compliant to File Hierarchy Standard.',
		'ELEM_COL_NAME' => 'File'
	},
	'RPM header check failed' => {
		'DESCRIPTION' => 'Some RPM header fields differ from what '.$appchecker_branch{'STD'}.' declares.',
		'ELEM_COL_NAME' => 'Element'
	},
	'RPM tags check failed' => {
		'DESCRIPTION' => 'Some RPM tags differ from what '.$appchecker_branch{'STD'}.' declares.',
		'ELEM_COL_NAME' => 'Element'
	},
	'RPM dependencies check failed' => {
		'DESCRIPTION' => 'RPM package includes dependencies not specified by '.$appchecker_branch{'STD'}.', or does not include required ones.',
		'ELEM_COL_NAME' => 'Dependency'
	},
	'RPM archive check failed' => {
		'DESCRIPTION' => 'Archive format of the RPM contents does not comply to '.$appchecker_branch{'STD'}.'.',
		'ELEM_COL_NAME' => 'Element'
	},
	'RPM install script not checked' => {
		'DESCRIPTION' => 'RPM install/uninstall scripts are not checked currently.',
		'ELEM_COL_NAME' => 'Element'
	},
	'RPM index check failed' => {
		'DESCRIPTION' => 'Problems encountered while checking RPM indices.',
		'ELEM_COL_NAME' => 'Element'
	},
	# Manual Tests
	'Manual Tests' => {
		'DESCRIPTION' => 'Tests which cannot be performed automatically.',
		'ELEM_COL_NAME' => 'Element'
	},
	# Automatic Tests (to show something if all automatic tests passed)
	'Automatic Tests' => {
		'DESCRIPTION' => 'All automatic tests finished successfully, no problems were detected.<br /><br />',
		'ELEM_COL_NAME' => 'Element'
	},
	# Uncategorized
	'Other' => {
		'DESCRIPTION' => 'All other problems not included in the categories given above.',
		'ELEM_COL_NAME' => 'Element'
	}
);

# Information about subcategories.
# %subcategories: SUBCAT_ID => {
#                     'CATEGORY' => <CAT-ID>,           # Shown in Category column (unique category ID)
#                     'ELEM_LINK' => <http-link>,       # This link + element name == full link to Navigator
#                     'MESSAGE' => <msg>                # Message shown prior to full details
#                 };
my %subcategories = (
	# Problems caused by test suite itself or some environment
	'J_FORMAT_ERROR' => {
		'CATEGORY' => 'Journal format error',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'The line is not of TETJ format:'
	},
	'J_SYNTAX_ERROR' => {
		'CATEGORY' => 'Journal syntax error',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'The line format does not match its code:'
	},
	'J_TCC_MIDDLE' => {
		'CATEGORY' => 'Journal format error',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'TCC Start entry appears in the middle of the journal:'
	},
	'J_TC_NOT_STARTED' => {
		'CATEGORY' => 'Journal format error',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Test Case End entry found when no test case seems to be running.'
	},
	'J_TC_NOT_FINISHED' => {
		'CATEGORY' => 'Journal format error',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Test case starts before the previous one has finished:'
	},
	'TS_CONTROLLER_MSG' => {
		'CATEGORY' => 'Test case controller error',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Test case controller reports a problem:'
	},
	'TS_ERROR' => {
		'CATEGORY' => 'Test suite error',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Error reported by the test suite:'
	},
	'R_EMPTY_TEST_CASE' => {
		'CATEGORY' => 'Empty test case',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'This test case contains no test points.'
	},
	'TS_ABORTED' => {
		'CATEGORY' => 'Test suite run aborted',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Test suite execution interrupted by the user.'
	},
	'J_TP_RESTARTED' => {
		'CATEGORY' => 'Test point restarted',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Test was restarted in the middle.'
	},
	'J_TP_NOT_STARTED' => {
		'CATEGORY' => 'Journal format error',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Test result entry found but the test point had not started.'
	},
	'J_TP_NOT_FINISHED' => {
		'CATEGORY' => 'Journal format error',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Test point starts before the previous one has finished:'
	},
	'J_TP_CHANGED' => {
		'CATEGORY' => 'Journal format error',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Test point number at the end is different from the starting one:'
	},
	'R_UNKNOWN_RESULT' => {
		'CATEGORY' => 'Unrecognized result code',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Unrecognized test result code.'
	},
	'R_FAILED_UNPACK' => {
		'CATEGORY' => 'Wrong archive format',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'The archive could not be unpacked:'
	},
	'R_NON_STD_ARCHIVE' => {
		'CATEGORY' => 'Wrong archive format',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'The archive is not '.$appchecker_branch{'STD'}.' compliant:'
	},
	# Problems found in the files tested
	# appchk-elf
	'R_ELF_HDR_FAILED' => {
		'CATEGORY' => 'ELF header check failed',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_LOAD_HDR_FAILED' => {
		'CATEGORY' => 'ELF header check failed',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_DYN_FAILED' => {
		'CATEGORY' => 'Dynamic segment check failed',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_WRONG_LOADER' => {
		'CATEGORY' => 'Incorrect program loader',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'For '.$appchecker_branch{'STD'}.' applications the program loader should be something like <b>/lib/ld-lsb.so.3</b>.<br />Messages from the test:'
	},
	'R_NOTE_FAILED' => {
		'CATEGORY' => 'NOTE section check failed',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_TLS_FAILED' => {
		'CATEGORY' => 'TLS section check failed',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_GNU_STACK_FAILED' => {
		'CATEGORY' => 'GNU_STACK header check failed',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_ELF_SECTION_FAILED' => {
		'CATEGORY' => 'ELF section check failed',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_ELF_SECTION_NON_STD' => {
		'CATEGORY' => 'Non-'.$appchecker_branch{'STD'}.' ELF section',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_LIBRARY_NON_STD' => {
		'CATEGORY' => 'Non-'.$appchecker_branch{'STD'}.' library used',
		'ELEM_LINK' => $appchecker_branch{'NAVIGATOR'}.'rawlib.php?cmd=list-byrunname&amp;RLrunname=',
		'MESSAGE' => 'Messages from the test:'
	},
	'R_INTERFACE_NON_STD' => {
		'CATEGORY' => 'Non-'.$appchecker_branch{'STD'}.' interface used',
		'ELEM_LINK' => $appchecker_branch{'NAVIGATOR'}.'int_single.php?cmd=list-by-name&amp;Iname=',
		'MESSAGE' => 'Messages from the test:'
	},
	'R_IO_FAILED' => {
		'CATEGORY' => 'I/O problem',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_INTERFACE_VERSION_FAILED' => {
		'CATEGORY' => 'Unexpected interface version',
		'ELEM_LINK' => $appchecker_branch{'NAVIGATOR'}.'int_single.php?cmd=list-by-name&amp;Iname=',
		'MESSAGE' => 'Messages from the test:'
	},
	'R_INTERFACE_DEPRECATED' => {
		'CATEGORY' => 'Deprecated interface used',
		'ELEM_LINK' => $appchecker_branch{'NAVIGATOR'}.'int_single.php?cmd=list-by-name&amp;Iname=',
		'MESSAGE' => 'Messages from the test:'
	},
	'R_IOCTL_FIP' => {
		'CATEGORY' => 'ioctl() encountered',
		'ELEM_LINK' => $appchecker_branch{'NAVIGATOR'}.'int_single.php?cmd=list-by-name&amp;Iname=',
		'MESSAGE' => 'Messages from the test:'
	},
	'R_DLOPEN_FIP' => {
		'CATEGORY' => 'dlopen() encountered',
		'ELEM_LINK' => $appchecker_branch{'NAVIGATOR'}.'int_single.php?cmd=list-by-name&amp;Iname=',
		'MESSAGE' => 'Messages from the test:'
	},
	# appchk-shell (Linux AppChecker only)
	'R_SH_BAD_SHELL' => {
		'CATEGORY' => 'Wrong shell script interpreter',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_SH_BASHISM' => {
		'CATEGORY' => 'Bashism used in shell script',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_SH_BAD_CMD' => {
		'CATEGORY' => 'Non-'.$appchecker_branch{'STD'}.' command used',
		'ELEM_LINK' => $appchecker_branch{'NAVIGATOR'}.'cmd_single.php?cmd=list-by-name&Cname=',
		'MESSAGE' => 'Messages from the test:'
	},
	'R_SH_BAD_COMPARISON' => {
		'CATEGORY' => 'Wrong comparison operator',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_SH_BAD_FUNC_DEF' => {
		'CATEGORY' => 'Wrong function definition',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_SH_BAD_ID_NAME' => {
		'CATEGORY' => 'Wrong identifier',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_SH_BAD_TILDE' => {
		'CATEGORY' => 'Wrong tilde expansion',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_SH_PARSE_ERROR' => {
		'CATEGORY' => 'Parse error',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_SH_AMBIGUOUS_SYNTAX' => {
		'CATEGORY' => 'Ambiguous syntax',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_SH_INCLUDE_ERROR' => {
		'CATEGORY' => 'Include error',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_SH_EXIT' => {
		'CATEGORY' => 'Unconditional exit',
		'ELEM_LINK' => $appchecker_branch{'NAVIGATOR'}.'cmd_single.php?cmd=list-by-name&Cname=',
		'MESSAGE' => 'Messages from the test:'
	},
	# appchk-perl
	'R_PERL_VERSION_FAILED' => {
		'CATEGORY' => 'Non-'.$appchecker_branch{'STD'}.' Perl version used',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_PERL_MODULE_NON_STD' => {
		'CATEGORY' => 'Non-'.$appchecker_branch{'STD'}.' Perl module used',
		'ELEM_LINK' => $appchecker_branch{'NAVIGATOR'}.'intlang.php?cmd=display-ilmodule-byname&amp;ILMlanguage=1&amp;ILMname=',
		'MESSAGE' => 'Messages from the test:'
	},
	# appchk-python
	'R_PYTHON_MODULE_PATH_NON_STD' => {
		'CATEGORY' => 'Non-'.$appchecker_branch{'STD'}.' Python module path used',
		'ELEM_LINK' => $appchecker_branch{'NAVIGATOR'}.'intlang.php?cmd=display-ilmodule-byname&amp;ILMlanguage=2&amp;ILMname=',
		'MESSAGE' => 'Messages from the test:'
	},
	'R_PYTHON_MODULE_NON_STD' => {
		'CATEGORY' => 'Non-'.$appchecker_branch{'STD'}.' Python module used',
		'ELEM_LINK' => $appchecker_branch{'NAVIGATOR'}.'intlang.php?cmd=display-ilmodule-byname&amp;ILMlanguage=2&amp;ILMname=',
		'MESSAGE' => 'Messages from the test:'
	},
	# desktopchk (Moblin AppChecker only)
	'R_DESKTOP_PATH_NON_STD' => {
		'CATEGORY' => 'Non-'.$appchecker_branch{'STD'}.' desktop path',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_DESKTOP_FIELD_NOT_FOUND' => {
		'CATEGORY' => 'Desktop field check failed',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_DESKTOP_ICON_NOT_FOUND' => {
		'CATEGORY' => 'Desktop icon check failed',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_DESKTOP_ICON_FORMAT_NON_STD' => {
		'CATEGORY' => 'Desktop icon check failed',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_DESKTOP_ICON_SIZE_FAILED' => {
		'CATEGORY' => 'Desktop icon check failed',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	# pkgchk (Moblin AppChecker only)
	'R_PKG_PATH_FAILED' => {
		'CATEGORY' => 'Path check failed',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	# pkgchk (Linux AppChecker only)
	'R_RPM_FHS_FAILED' => {
		'CATEGORY' => 'FHS check failed',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_RPM_HDR_FAILED' => {
		'CATEGORY' => 'RPM header check failed',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_RPM_TAG_FAILED' => {
		'CATEGORY' => 'RPM tags check failed',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_RPM_DEPS_FAILED' => {
		'CATEGORY' => 'RPM dependencies check failed',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_RPM_DEPS_FAILED_LINK' => {
		'CATEGORY' => 'RPM dependencies check failed',
		'ELEM_LINK' => $appchecker_branch{'NAVIGATOR'}.'rawlib.php?cmd=list-byrunname&amp;RLrunname=',
		'MESSAGE' => 'Messages from the test:'
	},
	'R_RPM_ARCHIVE_FAILED' => {
		'CATEGORY' => 'RPM archive check failed',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_RPM_SCRIPTS_UNTESTED' => {
		'CATEGORY' => 'RPM install script not checked',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	'R_RPM_IDX_FAILED' => {
		'CATEGORY' => 'RPM index check failed',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	},
	# Manual Tests
	'R_MANUAL' => {
		'CATEGORY' => 'Manual Tests',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Requirement:'
	},
	# Uncategorized
	'R_OTHER' => {
		'CATEGORY' => 'Other',
		'ELEM_LINK' => undef,
		'MESSAGE' => 'Messages from the test:'
	}
);

# Sorted list of categories
my @categories_sort = (
	# Problems caused by test suite itself or some environment
	'Journal format error',
	'Journal syntax error',
	'Test case controller error',
	'Test suite error',
	'Empty test case',
	'Test suite run aborted',
	'Test point restarted',
	'Unrecognized result code',
	'Wrong archive format',
	# Problems found in the files tested
	# appchk-elf
	'ELF header check failed',
	'Dynamic segment check failed',
	'Incorrect program loader',
	'NOTE section check failed',
	'TLS section check failed',
	'GNU_STACK header check failed',
	'ELF section check failed',
	'Non-'.$appchecker_branch{'STD'}.' ELF section',
	'Non-'.$appchecker_branch{'STD'}.' library used',
	'Non-'.$appchecker_branch{'STD'}.' interface used',
	'I/O problem',
	'Unexpected interface version',
	'Deprecated interface used',
	'ioctl() encountered',
	'dlopen() encountered',
	# appchk-shell (Linux AppChecker only)
	'Wrong shell script interpreter',
	'Bashism used in shell script',
	'Non-'.$appchecker_branch{'STD'}.' command used',
	'Wrong comparison operator',
	'Wrong function definition',
	'Wrong identifier',
	'Wrong tilde expansion',
	'Ambiguous syntax',
	'Include error',
	'Unconditional exit',
	'Parse error',
	# appchk-perl
	'Non-'.$appchecker_branch{'STD'}.' Perl version used',
	'Non-'.$appchecker_branch{'STD'}.' Perl module used',
	# appchk-python
	'Non-'.$appchecker_branch{'STD'}.' Python module path used',
	'Non-'.$appchecker_branch{'STD'}.' Python module used',
	# desktopchk (Moblin AppChecker only)
	'Non-'.$appchecker_branch{'STD'}.' desktop path',
	'Desktop field check failed',
	'Desktop icon check failed',
	# pkgchk (Moblin AppChecker only)
	'Path check failed',
	# pkgchk (Linux AppChecker only)
	'FHS check failed',
	'RPM header check failed',
	'RPM tags check failed',
	'RPM dependencies check failed',
	'RPM archive check failed',
	'RPM install script not checked',
	'RPM index check failed',
	# Manual Tests
	'Manual Tests',
	# Uncategorized
	'Other'
);

# Reversed hash: problems list for each category
# %cat_problems: <CAT-ID> => {
#                    'FAILS' => <number>            # Total number of fails in this category
#                    'WARNS' => <number>            # Total number of warnings in this category
#                    'FIPS' => <number>             # Total number of FIPs in this category
#                    'PROBLEMS' => [                # List of problems
#                        {
#                            'JOURNAL_TITLE' => <journal_title>     # Reference to journal info
#                            'PROBLEM_IDX' => <idx>                 # Index of the specific problem
#                        },
#                        ...
#                    ]
#                }
my %cat_problems = ();

# Reversed hash for retrieving journal title from HTML anchor.
my %journal_title_from_anchor;

# Map of TET result codes to human-readable severity names.
my %result_code_map = (
	0, "Passed",
	1, "Failed",
	2, "Unresolved",
	3, "Unused",
	4, "Unsupported",
	5, "Untested",
	6, "Uninitiated",
	7, "Unreported",
	65, "Time Expired",
	101, "Warning",
	102, "FIP",
	103, "Unimplemented",
	104, "Unapproved"
);

# Global statistics
my $tests_cancelled = 0;    # Number of cancelled tests

################################################################################
# Functions

# Prints usage info and exits.
sub usage() {
	print "Usage: $0 <test_journals_dir> [-o output-file.tgz] [-n app-name]\n";
	exit(1);
}

# Parses command line options passed to the script.
sub parse_cmd_line() {
	my $argc = scalar(@ARGV);
	if ($argc < 1) {
		usage();
	}

	for (my $i = 0; $i < $argc; ++$i) {
		if ($ARGV[$i] eq '-o') {
			usage() if ($i == $argc - 1);
			++$i;
			$result_archive = $ARGV[$i];
		}
		elsif ($ARGV[$i] eq '-n') {
			usage() if ($i == $argc - 1);
			++$i;
			$application_name = $ARGV[$i];
		}
		else {
			$result_path = abs_path($ARGV[$i]);
			$result_path =~ s!/+$!!;
			$result_path =~ m!^(.*)/([^/]+)$!;
			$result_path = ($1 or '/');
			$result_dir = $2;
		}
	}
	unless ($result_archive) {
		$result_archive = "$result_dir.tgz";
	}
}

#Reads information about host system
sub read_host_info() {
	$spec_version = read_file("$result_path/$result_dir/test_logs/spec_version");
	if ($spec_version eq '') {
		if ($appchecker_branch{'STD'} eq 'LSB') {
			$spec_version = (read_file("$result_path/$result_dir/test_logs/lsb_version") or '4.0');
		}
		else {
			$spec_version = (read_file("$result_path/$result_dir/test_logs/moblin_version") or '2.0');
		}
	}
	$spec_version =~ s/[\r\n]//g;

	$test_comments = read_file("$result_path/$result_dir/test_logs/test_comments");

	$host_machine  = `uname -m`;
	$host_machine  =~ s/[\r\n]//g;

	$host_name     = `uname -n`;
	$host_name     =~ s/[\r\n]//g;

	$test_arch     = (read_file("$result_path/$result_dir/test_logs/host_platform") or $host_machine);
	$test_arch     =~ s/[\r\n]//g;

	$host_os = `lsb_release -s -d`;
	$host_os =~ s/\"(.*)\"/$1/;	# remove quotes
	$host_os =~ s/[\r\n]//g;

	$host_kernel = `uname -r -v`;
	$host_kernel =~ s/[\r\n]//g;
}

#---------------------------------------------------------------------

# Escape HTML entities
sub html($) {
	my ($text) = @_;

	if (!defined($text)) {
		return '';
	}

	$text = esc_html($text);
	$text =~ s/\n/<br \/>/g;

	return $text;
}

# Escape HTML entities except for text in {noreplace}{/noreplace}
sub html_excl($) {
	my ($text) = @_;

	(defined($text)) or return undef;

	my $res = '';

	while ($text =~ s/^(.*?){noreplace}//s) {
		$res .= html($1);
		if ($text =~ s/^(.*?){\/noreplace}//s) {
			$res .= $1;     # without escaping
		}
		else {
			$res .= '{noreplace}';
		}
	}

	$res .= html($text);

	return $res;
}

#----------------------------------------------------------------------

# Reads all journals and generates report
sub generate_report() {
	print "Generating HTML report\n";

	# Go to the output directory
	chdir("$result_path/$result_dir");

	# Read the manual tests journal if present
	$manual_tests_data = new SimpleJournalParser("$result_path/$result_dir/test_journals/manual_tests.simplejournal");

	# Parse each test journal
	opendir(DIR, 'test_journals') or die "Can't open directory $result_path/$result_dir/test_journals: $!\n";
	my @files = readdir(DIR);
	closedir(DIR);
	foreach my $file (@files) {
		if ($file =~ m/^(.*\.INFO)$/) {
			parse_journal("test_journals/$file");
		}
	}

	# Add manual tests
	parse_manual_journal();

	# If application name was not specified, use all journal titles.
	if (!$application_name) {
		$application_name = join(', ', @journal_titles);
	}

	# Sum up statistics
	my $total_stats = sum_stats();

	# Give the final verdict
	my $overall_verdict;
	if (($total_stats->{'FAILS'} > 0) or ($tests_cancelled > 0)) {
		$overall_verdict = 'Failed';
	}
	elsif (($total_stats->{'WARNS'} > 0) or ($total_stats->{'FIPS'} > 0)) {
		$overall_verdict = 'Warnings';
	}
	elsif (($total_stats->{'TOTAL'} == 0)) {
		$overall_verdict = 'No result';
	}
	else {
		$overall_verdict = 'Success';
	}

	# Write INFO_STD file
	my $INFO_filename = "$result_path/$result_dir/INFO_STD";
	if (open(INFO, ">$INFO_filename")) {
		print INFO "Verdict: $overall_verdict\n";
		close(INFO);
	}
	else {
		print "Couldn't open file '$INFO_filename' for writing: $!";
	}

	compile_report();
}

# TREE: {<node_name> => {'type' => (file|dir|pkg), 'children' => TREE, 'journal_anchor' => <anchor>}};
my %journal_tree;

# Adds element to the branch of journal tree.
sub add_elem_to_subtree($$$$);  # Prototype for recursive function
sub add_elem_to_subtree($$$$) {
	my ($subtree, $elem_name, $f_type, $j_anchor) = @_;
	$elem_name =~ s/^\/+//;
	my $idx = index($elem_name, '/');
	if ($idx > 0) {
		my $d_name = substr($elem_name, 0, $idx);
		$elem_name = substr($elem_name, $idx + 1);
		if (!$subtree->{'children'}->{$d_name}->{'type'}) {
			$subtree->{'children'}->{$d_name}->{'type'} = 'dir';
		}
		add_elem_to_subtree($subtree->{'children'}->{$d_name}, $elem_name, $f_type, $j_anchor);
	}
	else {
		$subtree->{'children'}->{$elem_name}->{'type'} = $f_type;
		$subtree->{'children'}->{$elem_name}->{'journal_anchor'} = $j_anchor;
	}
}

# Adds element to the journal tree.
sub add_elem_to_tree($$$) {
	my ($title, $anchor, $file) = @_;
	my $type = ((substr($title, -1) eq '|') ? 'pkg' : 'file');
	my @tmp = split(/\|/, $title, 2);
	if ($tmp[1]) {
		$tmp[1] =~ s/\|//g;
		$title = $tmp[0].'|'.$tmp[1];
	}
	else {
		$title = $tmp[0];
	}

	if ($title =~ m/^[^|]+$/) {
   		$journal_tree{$title}->{'type'} = $type;
   		$journal_tree{$title}->{'journal_anchor'} = $anchor;
	}
	elsif ($title =~ m/^([^|]+)\|(.+)$/) {
		my ($tree_root, $elem_name) = ($1, $2);
   		if (!$journal_tree{$tree_root}) {
			if ($tree_root =~ m/^PKG: /) {
				$journal_tree{$tree_root}->{'type'} = 'pkg';
			}
			else {
				$journal_tree{$tree_root}->{'type'} = 'dir';
			}
   		}
   		add_elem_to_subtree($journal_tree{$tree_root}, $elem_name, $type, $anchor);
	}
	else {
		print "Internal error: wrong format for TITLE ($title).\nFile: $file\n";
	}
}

# Map from journal status code to statistics category
my %stats_map = (
	'passed' => 'PASSES',
	'fip: passed' => 'PASSES',
	'manual: passed' => 'PASSES',
	'failed' => 'FAILS',
	'fip: failed' => 'FAILS',
	'unresolved' => 'FAILS',
	'unused' => 'SKIPPED',
	'unsupported' => 'SKIPPED',
	'untested' => 'NOTICES',
	'uninitiated' => 'FAILS',
	'unreported' => 'FAILS',
	'time expired' => 'FAILS',
	'warning' => 'WARNS',
	'fip' => 'FIPS',
	'manual: not performed' => 'FIPS',
	'unimplemented' => 'SKIPPED',
	'unapproved' => 'SKIPPED',
	'test suite error' => 'FAILS',
	'notice' => 'NOTICES'
);

# Adds a new problem entry into parsed journal info
sub add_journal_entry($$$$$$$) {
	my ($journal_title_data, $severity, $test_case, $subcategory, $element, $comment, $details) = @_;
	my $lc_severity = lc($severity);
	my $stat_name = $stats_map{$lc_severity};
	if ($stat_name) {
		++$journal_title_data->{'STATISTICS'}->{$stat_name};
		if (($stat_name eq 'SKIPPED') or ($lc_severity eq 'passed')) {
			$severity = '';     # Ignore this entry
		}
		elsif ($lc_severity eq 'manual: passed') {
			$severity = 'Passed';
		}
		elsif ($lc_severity eq 'manual: not performed') {
			$severity = 'Not performed';
		}
	}
	else {
		$severity = "Unrecognized ($severity)";
	}
	if ($severity) {
		push @{$journal_title_data->{'PROBLEMS_LIST'}}, {
			'SEVERITY' => $severity,
			'TEST_CASE' => $test_case,
			'SUBCATEGORY' => $subcategory,
			'ELEMENT' => $element,
			'COMMENT' => $comment,
			'DETAILS' => $details
		};
	}
}

# Analyzes the test messages and returns an array of problem entries
sub guess_problem_subcategory($$) {
	my ($test_messages, $test_tool) = @_;
	if (!$test_messages) {
		return [{'SUBCATEGORY' => 'R_OTHER', 'ELEMENT' => '', 'MESSAGE' => ''}];
	}
	my @msgs = split(/\n/, $test_messages);
	my $res = [];
	my $current = {'SUBCATEGORY' => '', 'ELEMENT' => '', 'MESSAGE' => ''};
	foreach my $msg (@msgs) {
		my $tmp = guess_problem_subcategory_line($msg, $test_tool);
		next if (!defined $tmp);
		if (($tmp->{'SUBCATEGORY'} eq 'R_LIBRARY_NON_STD') or
			($tmp->{'SUBCATEGORY'} eq 'R_RPM_DEPS_FAILED') or
			($tmp->{'SUBCATEGORY'} eq 'R_RPM_DEPS_FAILED_LINK') or
			($tmp->{'SUBCATEGORY'} eq 'R_DYN_FAILED')
			) {
			if ($current->{'SUBCATEGORY'}) {
				push @$res, $current;
				$current = {'SUBCATEGORY' => '', 'ELEMENT' => '', 'MESSAGE' => ''};
			}
			push @$res, {'SUBCATEGORY' => $tmp->{'SUBCATEGORY'}, 'ELEMENT' => $tmp->{'ELEMENT'}, 'MESSAGE' => $msg};
		}
		else {
			$current->{'MESSAGE'} .= ($current->{'MESSAGE'} ? "\n" : '').$msg;
			if (!$current->{'SUBCATEGORY'}) {
				$current->{'SUBCATEGORY'} = $tmp->{'SUBCATEGORY'};
				$current->{'ELEMENT'} = $tmp->{'ELEMENT'};
			}
			else {
				if (!$current->{'ELEMENT'}) {
					$current->{'ELEMENT'} = $tmp->{'ELEMENT'};
				}
				elsif ($current->{'ELEMENT'} ne $tmp->{'ELEMENT'}) {
					$current->{'ELEMENT'} = '&hellip;';
				}
			}
		}
	}
	if ($current->{'SUBCATEGORY'}) {
		push @$res, $current;
	}
	return $res;
}

# Analyzes the test message line and determines its subcategory and significant element
sub guess_problem_subcategory_line($$) {
	my ($msg_line, $test_tool) = @_;
	if (!$msg_line) {
		return {'SUBCATEGORY' => 'R_OTHER', 'ELEMENT' => ''};
	}
	my $res;
	if ($msg_line =~ m/Failed to unpack the archive\./i) {
		$res = {'SUBCATEGORY' => 'R_FAILED_UNPACK', 'ELEMENT' => ''};
	}
	elsif ($msg_line =~ m/No contents found in the DEB archive\./i) {
		$res = {'SUBCATEGORY' => 'R_FAILED_UNPACK', 'ELEMENT' => ''};
	}
	elsif ($msg_line =~ m/.* archives are not standardized by the $appchecker_branch{'STD'}\./i) {
		$res = {'SUBCATEGORY' => 'R_NON_STD_ARCHIVE', 'ELEMENT' => ''};
	}
	# appchk-elf
	elsif ($msg_line =~ m/compareElfhdr: (.*) isn\'t expected value /i) {
		$res = {'SUBCATEGORY' => 'R_ELF_HDR_FAILED', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/compareElfhdr: (.*) is neither /i) {
		$res = {'SUBCATEGORY' => 'R_ELF_HDR_FAILED', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/Section (.*) does not fit in Segment/i) {
		$res = {'SUBCATEGORY' => 'R_LOAD_HDR_FAILED', 'ELEMENT' => "Section $1"};
	}
	elsif ($msg_line =~ m/Section (.*) flags .* does not correspond to Segment flags /i) {
		$res = {'SUBCATEGORY' => 'R_LOAD_HDR_FAILED', 'ELEMENT' => "Section $1"};
	}
	elsif ($msg_line =~ m/Dynamic section (.*) does not match Segment /i) {
		$res = {'SUBCATEGORY' => 'R_DYN_FAILED', 'ELEMENT' => "Section $1"};
	}
	elsif ($msg_line =~ m/Non-dynamic section (.*) found in PT_DYNAMIC segment/i) {
		$res = {'SUBCATEGORY' => 'R_DYN_FAILED', 'ELEMENT' => "Section $1"};
	}
	elsif ($msg_line =~ m/Incorrect program interpreter: (.*), expected:/i) {
		$res = {'SUBCATEGORY' => 'R_WRONG_LOADER', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/NOTE (section .*) does not match Segment /i) {
		$res = {'SUBCATEGORY' => 'R_NOTE_FAILED', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/Section (.*) in PT_NOTE segment is not type SHT_NOTE/i) {
		$res = {'SUBCATEGORY' => 'R_NOTE_FAILED', 'ELEMENT' => "Section $1"};
	}
	elsif ($msg_line =~ m/PT_SHLIB is not permitted!/i) {
		$res = {'SUBCATEGORY' => 'R_OTHER', 'ELEMENT' => ''};
	}
	elsif ($msg_line =~ m/PHDR offset doesn\'t match expected /i) {
		$res = {'SUBCATEGORY' => 'R_OTHER', 'ELEMENT' => ''};
	}
	elsif ($msg_line =~ m/TLS (section .*) does not match Segment /i) {
		$res = {'SUBCATEGORY' => 'R_TLS_FAILED', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/Section (.*) in PT_TLS segment is not type SHF_TLS/i) {
		$res = {'SUBCATEGORY' => 'R_TLS_FAILED', 'ELEMENT' => "Section $1"};
	}
	elsif ($msg_line =~ m/Section in PT_GNU_EH_FRAME segment is not ".eh_frame_hdr"/i) {
		$res = {'SUBCATEGORY' => 'R_OTHER', 'ELEMENT' => ''};
	}
	elsif ($msg_line =~ m/(offset|vaddr|paddr|filesz|memsz) not 0/i) {
		$res = {'SUBCATEGORY' => 'R_GNU_STACK_FAILED', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/Unexpected flags /i) {
		$res = {'SUBCATEGORY' => 'R_GNU_STACK_FAILED', 'ELEMENT' => 'flags'};
	}
	elsif ($msg_line =~ m/IA_64_UNWIND not checked!/i) {
		$res = {'SUBCATEGORY' => 'R_OTHER', 'ELEMENT' => ''};
	}
	elsif ($msg_line =~ m/Section (.*): .* is wrong\. /i) {
		$res = {'SUBCATEGORY' => 'R_ELF_SECTION_FAILED', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/Section (.*) not checked/i) {
		$res = {'SUBCATEGORY' => 'R_ELF_SECTION_FAILED', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/section (.*) is not in the $appchecker_branch{'STD'}/i) {
		$res = {'SUBCATEGORY' => 'R_ELF_SECTION_NON_STD', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/ignoring (.*): Not in the $appchecker_branch{'STD'} but not loaded/i) {
		$res = {'SUBCATEGORY' => 'R_ELF_SECTION_NON_STD', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/DT_NEEDED: (.*) is used, but not part of the $appchecker_branch{'STD'}/i) {
		$res = {'SUBCATEGORY' => 'R_LIBRARY_NON_STD', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/Dynamic Tag .* unknown/i) {
		$res = {'SUBCATEGORY' => 'R_OTHER', 'ELEMENT' => 'Dynamic Tag'};
	}
	elsif ($msg_line =~ m/Could not open (.*)/i) {
		$res = {'SUBCATEGORY' => 'R_IO_FAILED', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/Could not stat file/i) {
		$res = {'SUBCATEGORY' => 'R_IO_FAILED', 'ELEMENT' => ''};
	}
	elsif ($msg_line =~ m/Symbol (.*) is used, but is not included in /i) {
		$res = {'SUBCATEGORY' => 'R_INTERFACE_NON_STD', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/Symbol (.*) has version .* expecting /i) {
		$res = {'SUBCATEGORY' => 'R_INTERFACE_VERSION_FAILED', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/Symbol (.*) has been deprecated/i) {
		$res = {'SUBCATEGORY' => 'R_INTERFACE_DEPRECATED', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/Unable to determine if parameters to ioctl\(\) are used in accordance with the $appchecker_branch{'STD'}/i) {
		$res = {'SUBCATEGORY' => 'R_IOCTL_FIP', 'ELEMENT' => 'ioctl'};
	}
	elsif ($msg_line =~ m/Unable to automatically check libraries accessed via dlopen\(\)/i) {
		$res = {'SUBCATEGORY' => 'R_DLOPEN_FIP', 'ELEMENT' => 'dlopen'};
	}
	# appchk-shell (Linux AppChecker only)
	elsif ($msg_line =~ m/The '(.*)' shell is not included in $appchecker_branch{'STD'}\. You should port the script to the 'sh' language\./i) {
		$res = {'SUBCATEGORY' => 'R_SH_BAD_SHELL', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/The '(.*)' is not a shell interpreter\./i) {
		$res = {'SUBCATEGORY' => 'R_SH_BAD_SHELL', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/Probably should be '#!\/bin\/(.*)' here\./i) {
		$res = {'SUBCATEGORY' => 'R_SH_BAD_SHELL', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/'\+=' is a bashism\. Should be /i) {
		$res = {'SUBCATEGORY' => 'R_SH_BASHISM', 'ELEMENT' => 'String operator +='};
	}
	elsif ($msg_line =~ m/'(.*)' is a Bash built in command and should not be used in a 'sh' script\./i) {
		$res = {'SUBCATEGORY' => 'R_SH_BASHISM', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/'(.*)' isn\'t included in $appchecker_branch{'STD'} /i) {
		$res = {'SUBCATEGORY' => 'R_SH_BAD_CMD', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/The 'function' keyword in the definition of function '(.*)' should be omitted/i) {
		$res = {'SUBCATEGORY' => 'R_SH_BAD_FUNC_DEF', 'ELEMENT' => "function $1"};
	}
	elsif ($msg_line =~ m/'select' is a bashism. Don\'t use it\./i) {
		$res = {'SUBCATEGORY' => 'R_SH_BASHISM', 'ELEMENT' => '\'select\' command'};
	}
	elsif ($msg_line =~ m/'&>' is a bashism\. Should be '>filename 2>&1'/i) {
		$res = {'SUBCATEGORY' => 'R_SH_BASHISM', 'ELEMENT' => 'Redirection operator &>'};
	}
	elsif ($msg_line =~ m/Here-string syntax \('<<<'\) is a bashism\./i) {
		$res = {'SUBCATEGORY' => 'R_SH_BASHISM', 'ELEMENT' => 'Here-string operator <<<'};
	}
	elsif ($msg_line =~ m/Process substitution \('<\(list\)'\) is a bashism\./i) {
		$res = {'SUBCATEGORY' => 'R_SH_BASHISM', 'ELEMENT' => 'Process substitution \'<(list)\''};
	}
	elsif ($msg_line =~ m/NAME token is expected: '(.*)'\./i) {
		$res = {'SUBCATEGORY' => 'R_SH_BAD_ID_NAME', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/It would be better to escape the '\$' character in this case: '(.*)'\./i) {
		$res = {'SUBCATEGORY' => 'R_SH_AMBIGUOUS_SYNTAX', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/The '(.*)' variable has special meaning in bash and may be not set in sh\. /i) {
		$res = {'SUBCATEGORY' => 'R_SH_BASHISM', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/Use 'id -u' instead of \$EUID\./i) {
		$res = {'SUBCATEGORY' => 'R_SH_BASHISM', 'ELEMENT' => '$EUID'};
	}
	elsif ($msg_line =~ m/Arrays are not portable\./i) {
		$res = {'SUBCATEGORY' => 'R_SH_BASHISM', 'ELEMENT' => 'Array'};
	}
	elsif ($msg_line =~ m/'(.*)': Unportable parameter expansion/i) {
		$res = {'SUBCATEGORY' => 'R_SH_BASHISM', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/'(.*)' - indirect expansion isn\'t portable\./i) {
		$res = {'SUBCATEGORY' => 'R_SH_BASHISM', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/Wrong tilde expansion: '(.*)'\. Only '~' and '~login' are supported\./i) {
		$res = {'SUBCATEGORY' => 'R_SH_BAD_TILDE', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/'\$\(< file\)' is a bashism\. Should be /i) {
		$res = {'SUBCATEGORY' => 'R_SH_BASHISM', 'ELEMENT' => 'Process substitution \'$(< file)\''};
	}
	elsif ($msg_line =~ m/\[\[\.\.\.\]\] syntax is a bashism and not portable\./i) {
		$res = {'SUBCATEGORY' => 'R_SH_BASHISM', 'ELEMENT' => '[[...]]'};
	}
	elsif ($msg_line =~ m/\(\(\.\.\.\)\) syntax is a bashism. Use \$\(\(\.\.\.\)\) instead, although note that this substitutes the value of the expression\./i) {
		$res = {'SUBCATEGORY' => 'R_SH_BASHISM', 'ELEMENT' => '((...))'};
	}
	elsif ($msg_line =~ m/'(.* \(\( \.\.\. \)\))' syntax is a bashism and not portable\./i) {
		$res = {'SUBCATEGORY' => 'R_SH_BASHISM', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/In arguments of the 'test' command use '=' rather than '=='\./i) {
		$res = {'SUBCATEGORY' => 'R_SH_BAD_COMPARISON', 'ELEMENT' => '=='};
	}
	elsif ($msg_line =~ m/'(.*)': Pattern replacement is a bashism. Use 'sed' instead\./i) {
		$res = {'SUBCATEGORY' => 'R_SH_BASHISM', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/ANSI-C Quoting syntax is not portable: '\$.*'\./i) {
		$res = {'SUBCATEGORY' => 'R_SH_BASHISM', 'ELEMENT' => '$\'...\''};
	}
	elsif ($msg_line =~ m/Locale Translation syntax is not portable: '\$.*'\./i) {
		$res = {'SUBCATEGORY' => 'R_SH_BASHISM', 'ELEMENT' => '$"..."'};
	}
	elsif ($msg_line =~ m/According to POSIX the part of redirection in line '.*' should be treated as '(.*)'\./i) {
		$res = {'SUBCATEGORY' => 'R_SH_AMBIGUOUS_SYNTAX', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/\$\[\.\.\.\] syntax is not portable\. /i) {
		$res = {'SUBCATEGORY' => 'R_SH_BASHISM', 'ELEMENT' => '$[...]'};
	}
	elsif ($msg_line =~ m/Comment line ends with backslash\. /i) {
		$res = {'SUBCATEGORY' => 'R_SH_AMBIGUOUS_SYNTAX', 'ELEMENT' => '# \\'};
	}
	elsif ($msg_line =~ m/According to the standard 'for .*' should be replaced with 'for .*' \(semicolon omited\)\./i) {
		$res = {'SUBCATEGORY' => 'R_SH_BASHISM', 'ELEMENT' => 'for ...; do'};
	}
	elsif ($msg_line =~ m/Parsing error at /i) {
		$res = {'SUBCATEGORY' => 'R_SH_PARSE_ERROR', 'ELEMENT' => ''};
	}
	elsif ($msg_line =~ m/Unknown file was included via the dot operator: '\. (.*)'$/i) {
		$res = {'SUBCATEGORY' => 'R_SH_INCLUDE_ERROR', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/Possible loop including\./i) {
		$res = {'SUBCATEGORY' => 'R_SH_PARSE_ERROR', 'ELEMENT' => ''};
	}
	elsif ($msg_line =~ m/Unconditional exit faced( \(exec\))?. Stop parsing\./i) {
		$res = {'SUBCATEGORY' => 'R_SH_EXIT', 'ELEMENT' => ($1 ? 'exec' : 'exit')};
	}
	elsif ($msg_line =~ m/CRLF linebreak isn\'t supported by sh\./) {
		$res = {'SUBCATEGORY' => 'R_SH_PARSE_ERROR', 'ELEMENT' => 'CRLF'};
	}
	elsif ($msg_line =~ m/Extended patterns are not portable\./) {
		$res = {'SUBCATEGORY' => 'R_SH_BASHISM', 'ELEMENT' => 'case'};
	}
	elsif ($msg_line =~ m/\$_ is not described in the shell language standard/) {
		$res = {'SUBCATEGORY' => 'R_SH_BASHISM', 'ELEMENT' => '$_'};
	}
	elsif ($msg_line =~ m/Bash brace expansion syntax '{a,b,c}' is not portable: '(.*)'/) {
		$res = {'SUBCATEGORY' => 'R_SH_BASHISM', 'ELEMENT' => 'a{b,c}d'};
	}
	# appchk-perl & appchk-python
	elsif ($msg_line =~ m/Requires perl version (.*) but $appchecker_branch{'STD'} specifies /i) {
		$res = {'SUBCATEGORY' => 'R_PERL_VERSION_FAILED', 'ELEMENT' => "Perl $1"};
	}
	elsif ($msg_line =~ m/(.*) is used, but is not part of $appchecker_branch{'STD'}/i) {
		if ($test_tool =~ m/-appchk-perl$/) {
			$res = {'SUBCATEGORY' => 'R_PERL_MODULE_NON_STD', 'ELEMENT' => $1};
		}
		else {
			$res = {'SUBCATEGORY' => 'R_PYTHON_MODULE_NON_STD', 'ELEMENT' => $1};
		}
	}
	elsif ($msg_line =~ m/(.*) is used from nonstandard path /i) {
		$res = {'SUBCATEGORY' => 'R_PYTHON_MODULE_PATH_NON_STD', 'ELEMENT' => $1};
	}
	# desktopchk (Moblin AppChecker only)
	elsif ($msg_line =~ m/library version is (.*) than $appchecker_branch{'STD'} compliant library version/i) {
		$res = {'SUBCATEGORY' => 'R_DESKTOP_PATH_NON_STD', 'ELEMENT' => ''};
	}
	elsif ($msg_line =~ m/field \"(.*)\" doesn\'t exist/i) {
		$res = {'SUBCATEGORY' => 'R_DESKTOP_FIELD_NOT_FOUND', 'ELEMENT' => ''};
	}
	elsif ($msg_line =~ m/field \"(.*)\" or \"(.*)\" doesn\'t exist/i) {
		$res = {'SUBCATEGORY' => 'R_DESKTOP_FIELD_NOT_FOUND', 'ELEMENT' => ''};
	}
	elsif ($msg_line =~ m/unable to find specified icon: (.*)/i) {
		$res = {'SUBCATEGORY' => 'R_DESKTOP_ICON_NOT_FOUND', 'ELEMENT' => ''};
	}
	elsif ($msg_line =~ m/icon file found: (.*), and it\'s not $appchecker_branch{'STD'} compliant/i) {
		$res = {'SUBCATEGORY' => 'R_DESKTOP_ICON_NOT_FOUND', 'ELEMENT' => ''};
	}
	elsif ($msg_line =~ m/icon file found: (.*), size not found:/i) {
		$res = {'SUBCATEGORY' => 'R_DESKTOP_ICON_SIZE_FAILED', 'ELEMENT' => ''};
	}
	elsif ($msg_line =~ m/Incompliant icon path '(?:.*\/)?([^\/]*)'/i) {
		$res = {'SUBCATEGORY' => 'R_PKG_PATH_FAILED', 'ELEMENT' => $1};
	}
	# pkgchk (Moblin AppChecker only)
	elsif ($msg_line =~ m/Location of the file '(.*)' is not $appchecker_branch{'STD'} compliant/i) {
		$res = {'SUBCATEGORY' => 'R_PKG_PATH_FAILED', 'ELEMENT' => $1};
	}
	# pkgchk (Linux AppChecker only)
	elsif ($msg_line =~ m/checkRpmArchive: /i) {
		$res = {'SUBCATEGORY' => 'R_RPM_HDR_FAILED', 'ELEMENT' => ''};
	}
	elsif ($msg_line =~ m/compareRpmHeader: /i) {
		$res = {'SUBCATEGORY' => 'R_RPM_HDR_FAILED', 'ELEMENT' => ''};
	}
	elsif ($msg_line =~ m/compareRpmLead: /i) {
		$res = {'SUBCATEGORY' => 'R_RPM_HDR_FAILED', 'ELEMENT' => ''};
	}
	elsif ($msg_line =~ m/file (.*) not FHS compliant/i) {
		$res = {'SUBCATEGORY' => 'R_RPM_FHS_FAILED', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/(Type|Count) for Index (.*) does not match. /i) {
		$res = {'SUBCATEGORY' => 'R_RPM_TAG_FAILED', 'ELEMENT' => $2};
	}
	elsif ($msg_line =~ m/Incorrect (RPMTAG_OS|RPMTAG_ARCH): /i) {
		$res = {'SUBCATEGORY' => 'R_RPM_TAG_FAILED', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/check_dependencies\(\) different number of /i) {
		$res = {'SUBCATEGORY' => 'R_RPM_DEPS_FAILED', 'ELEMENT' => ''};
	}
	elsif ($msg_line =~ m/Version for (.*): .* doesn\'t match expected /i) {
		$res = {'SUBCATEGORY' => 'R_RPM_DEPS_FAILED', 'ELEMENT' => $1};
	}
	elsif ($msg_line =~ m/Unexpected dependency (.*)/i) {
		my $elem = $1;
		if ($elem =~ m/^([^\(]+\.so\.\d+)/) {
			$res = {'SUBCATEGORY' => 'R_RPM_DEPS_FAILED_LINK', 'ELEMENT' => $1};
		}
		else {
			$res = {'SUBCATEGORY' => 'R_RPM_DEPS_FAILED', 'ELEMENT' => $elem};
		}
	}
	elsif ($msg_line =~ m/Didn\'t see (required|expected) dependency (.*)/i) {
		$res = {'SUBCATEGORY' => 'R_RPM_DEPS_FAILED', 'ELEMENT' => $2};
	}
	elsif ($msg_line =~ m/Provide Flag not checked: /i) {
		$res = {'SUBCATEGORY' => 'R_OTHER', 'ELEMENT' => ''};
	}
	elsif ($msg_line =~ m/Optflags not checked: /i) {
		$res = {'SUBCATEGORY' => 'R_OTHER', 'ELEMENT' => ''};
	}
	elsif ($msg_line =~ m/Invalid RPMTAG_PAYLOADFORMAT: expecting cpio but found (.*)/i) {
		$res = {'SUBCATEGORY' => 'R_RPM_ARCHIVE_FAILED', 'ELEMENT' => "Format $1"};
	}
	elsif ($msg_line =~ m/Invalid RPMTAG_PAYLOADCOMPRESSOR: expecting gzip but found (.*)/i) {
		$res = {'SUBCATEGORY' => 'R_RPM_ARCHIVE_FAILED', 'ELEMENT' => "Compressor $1"};
	}
	elsif ($msg_line =~ m/Invalid RPMTAG_PAYLOADFLAGS: expecting "9" but found (.*)/i) {
		$res = {'SUBCATEGORY' => 'R_RPM_ARCHIVE_FAILED', 'ELEMENT' => "Flags $1"};
	}
	elsif ($msg_line =~ m/((Pre|Post)-(un)?install) program not checked/i) {
		# Testing scripts is performed separately, supporess the message.
		# $res = {'SUBCATEGORY' => 'R_RPM_SCRIPTS_UNTESTED', 'ELEMENT' => "$1 script"};
	}
	elsif ($msg_line =~ m/checkRpmIdx\(\) unexpected Index /i) {
		$res = {'SUBCATEGORY' => 'R_RPM_TAG_FAILED', 'ELEMENT' => ''};
	}
	elsif ($msg_line =~ m/checkRpmIdx\(\) .* Index (.*) (not )?found/i) {
		$res = {'SUBCATEGORY' => 'R_RPM_IDX_FAILED', 'ELEMENT' => $1};
	}
	else {
		# Uncategorized
		$res = {'SUBCATEGORY' => 'R_OTHER', 'ELEMENT' => ''};
	}

	return $res;
}

# Reads and parses specific TET journal specified by the INFO file.
sub parse_journal($) {
	my ($INFO_filename) = @_;
	my $journal_file;
	my $journal_title;
	my $journal_anchor;

	# Parse INFO file
	if (open(FILE, $INFO_filename)) {
		while (my $line = <FILE>) {
			if ($line =~ m/^\s*TITLE:\s+(.*)$/) {
				$journal_title = $1;
			}
			elsif ($line =~ m/^\s*ANCHOR:\s+(.*)$/) {
				$journal_anchor = $1;
			}
			elsif ($line =~ m/^\s*JOURNAL_FILENAME:\s+(.*)$/) {
				$journal_file = "test_journals/$1";
			}
		}
		close FILE;
	}
	else {
		print "Couldn't open file '$INFO_filename' for reading: $!";
	}
	if (!$journal_file) {
		$INFO_filename =~ m/^(.*)\.INFO$/;
		$journal_file = $1;
	}
	if (!$journal_anchor) {
		$journal_anchor = $journal_file;
		$journal_anchor =~ s/.*\///;    # Cut path
		$journal_anchor =~ s/\..*$//;   # Cut extension
	}
	if (!$journal_title) {
		$journal_title = $journal_anchor;
	}

	add_elem_to_tree($journal_title, $journal_anchor, $journal_file);

	# Store the journal title and anchor.
	push @journal_titles, $journal_title;
	$journal_title_from_anchor{$journal_anchor} = $journal_title;

	$journal_data{$journal_title} = {
		'JOURNAL_FILE' => $journal_file,
		'TEST_TOOL' => '&lt;Unknown&gt;',
		'START_TIME' => '&lt;Unknown&gt;',
		'PROBLEMS_LIST' => [],
		'IS_RESULT' => 1,
		'STATISTICS' => {
			'PASSES' => 0,
			'FAILS' => 0,
			'WARNS' => 0,
			'FIPS' => 0,
			'NOTICES' => 0,
			'QUELLED' => 0,
			'SKIPPED' => 0
		}
	};

	# Open the journal file.
	print "Reading journal: $journal_file\n";
	if (!open(JOURNAL, $journal_file)) {
		# The journal file could not be open; mention this in the log
		# (so that we know that something was missing) and return.

		my $journal_title_disp = $journal_title;
		$journal_title_disp =~ s/^\|//;
		$journal_title_disp =~ s/\|$//;
		$journal_title_disp =~ s/\|/ =&gt; /g;

		++$tests_cancelled;

		print "Failed to read journal: $journal_file.\n$journal_title_disp failed to run, interrupted or crashed during the execution.\n";
		$journal_data{$journal_title}->{'IS_RESULT'} = undef;
		return;
	}

	# This flag is set to nonzero if we are inside a test case.
	my $inside_test = 0;

	# Variables for storing information about test cases
	my $test_name;
	my $test_number;
	my $test_point;
	my $test_points_in_this_case;
	my $test_messages;
	my $line_number = 0;
	my $journal_empty = 1;

	# List of problems in the current journal
	my $journal_title_data = $journal_data{$journal_title};

	my %fip_details = (
		'R_IOCTL_FIP' => {
			'TITLE' => 'ioctl() calls',
			'DESCRIPTION' => 'The <b>ioctl()</b> interface is used by the application. Static analysis is unable to check if its parameters are used in accordance with the '.$appchecker_branch{'STD'}.'.'
		},
		'R_DLOPEN_FIP' => {
			'TITLE' => 'dlopen() calls',
			'DESCRIPTION' => 'The <b>dlopen()</b> interface is used to dynamically load shared objects at run-time. For application to be '.$appchecker_branch{'STD'}.' compliant, it is necessary that no libraries outside '.$appchecker_branch{'STD'}.' are used. Static analysis is unable to check this.'
		}
	);

	# Information about FIPs
	# ->{<CATEGORY>}{TITLE} = <value>
	#               {TESTCASES}{<ID>} = <message>
	my %fips = ();

	# Number of FIPs not yet covered by the user via Manual Tests
	my $remaining_fips = 0;

	my $test_file_name = (($journal_title =~ m!/([^/]+)$!) ? $1 : $journal_title);
	my $test_file_id = $journal_file;
	$test_file_id =~ s/^.*($test_file_name\.\d+-\d+)\.journal$/$1/;

	# Read the journal line by line.
	while (my $line = <JOURNAL>) {
		++$line_number;

		# Each journal line consists of three parts, separated by
		# a vertical bar:
		#
		# 1. A numeric line code, determining the meaning of the line.
		# 2. Additional data, format depends on line code.
		# 3. Additional text (message for the user, or some comment).
		unless ($line =~ /^(\d*)\|([^|]*)\|(.*)$/) {
			# Lines that do not conform to this format are reported
			# to the user and skipped.
			add_journal_entry(
				$journal_title_data,
				'Warning',
				'',
				'J_FORMAT_ERROR',
				"$journal_file:$line_number",
				'',
				$line
			);
			next;
		}
		my $line_code = $1;
		my $line_data = $2;
		my $line_text = $3;

		# Determine the type of the line and process it appropriately.
		if ($line_code == 0) {
			# Code 0 -- Test Case Controller Start.
			# The line structure is:
			# 0|version time date|who
			unless ($line_data =~ /^(\S*)\s*(\S*)\s(\S*)$/) {
				add_journal_entry(
					$journal_title_data,
					'Warning',
					'',
					'J_SYNTAX_ERROR',
					"$journal_file:$line_number",
					'',
					$line
				);
				next;
			}

			# This line should be one of the first;
			# if it comes in the middle of normal results, it is strange.
			if ($inside_test) {
				add_journal_entry(
					$journal_title_data,
					'Warning',
					'',
					'J_TCC_MIDDLE',
					"$journal_file:$line_number",
					'',
					"$line_data\n$line_text"
				);
			}
			else {
				# my $test_engine_version = $1;
				my $test_start_time = $2;
				my $test_start_date = $3;
				if ($test_start_date =~ m/^(\d{4})(\d{2})(\d{2})$/) {
					$test_start_date = "$3.$2.$1";
				}
				$journal_title_data->{'START_TIME'} = "$test_start_date $test_start_time";

				$line_text =~ s/^.*Command line:\s*//;
				if ($line_text =~ m!^\S*/([^\s/]+)\s?!) {
					$line_text = $1;
				}
				# Linux AppChecker tools
				if ($line_text eq 'lsbappchk') {
					$journal_title_data->{'TEST_TOOL'} = 'lsb-appchk';
				}
				elsif ($line_text eq 'lsbappchk.pl') {
					$journal_title_data->{'TEST_TOOL'} = 'lsb-appchk-perl';
				}
				elsif ($line_text eq 'lsbappchk.py') {
					$journal_title_data->{'TEST_TOOL'} = 'lsb-appchk-python';
				}
				elsif ($line_text eq 'lsbappchk-sh.pl') {
					$journal_title_data->{'TEST_TOOL'} = 'lsb-appchk-shell';
				}
				elsif ($line_text eq 'lsbpkgchk') {
					$journal_title_data->{'TEST_TOOL'} = 'lsb-pkgchk';
				}
				# Moblin AppChecker tools
				elsif ($line_text eq 'moblinappchk') {
					$journal_title_data->{'TEST_TOOL'} = 'moblin-appchk';
				}
				elsif ($line_text eq 'moblinappchk-perl.pl') {
					$journal_title_data->{'TEST_TOOL'} = 'moblin-appchk-perl';
				}
				elsif ($line_text eq 'moblinappchk-python.py') {
					$journal_title_data->{'TEST_TOOL'} = 'moblin-appchk-python';
				}
				elsif ($line_text eq 'moblindesktopchk.py') {
					$journal_title_data->{'TEST_TOOL'} = 'moblin-desktopchk';
				}
				elsif ($line_text eq 'moblinpkgchk.pl') {
					$journal_title_data->{'TEST_TOOL'} = 'moblin-pkgchk';
				}
				# Manual tests
				elsif ($line_text eq '<none>') {
					$journal_title_data->{'TEST_TOOL'} = 'None';
				}
				# Test tool is not specified
				else {
					$journal_title_data->{'TEST_TOOL'} = "Unknown ($line_text)";
				}
			}
		}
		elsif ($line_code == 5) {
			# Code 5 -- Local System Information.
			# Ignore this data - it's the same for all journals.
		}
		elsif ($line_code == 10) {
			# Code 10 -- Test Case Start.
			# The line structure is:
			# 10|activitynum testname hh:mm:ss|list of invocable components
			unless ($line =~ /^\d*\|(\d*)\s*(.*)\s+\S+\|/) {
				add_journal_entry(
					$journal_title_data,
					'Warning',
					'',
					'J_SYNTAX_ERROR',
					"$journal_file:$line_number",
					'',
					$line
				);
				next;
			}

			if ($test_name) {
				# The previous test case was not properly finished;
				# issue a warning.
				add_journal_entry(
					$journal_title_data,
					'Warning',
					"$test_name:0",
					'J_TC_NOT_FINISHED',
					"$journal_file:$line_number",
					'',
					"Previous test case name: $test_name"
				);
			}

			# Store the test name and number.
			$test_number = $1;
			$test_name = $2;
			$test_name =~ s|^.*/([^/]+)$|$1|;
			$test_name =~ s|^(.*)-pass2$|$1|;
			$inside_test = 1;

			# Reset the test point number and counter.
			$test_point = 0;
			$test_points_in_this_case = 0;
		}
		elsif ($line_code == 15) {
			# Code 15 -- Test Case Manager Start.
			# Not interesting to us.
		}
		elsif ($line_code == 20) {
			# Code 20 -- Local/Remote System Configuration Start.
			# Not interesting to us.
		}
		elsif ($line_code == 30) {
			# Code 30 -- Configuration Variables.
			# Not interesting to us.
		}
		elsif ($line_code == 40) {
			# Code 40 -- Configuration End.
			# Not interesting to us.
		}
		elsif ($line_code == 50) {
			# Code 50 -- Test Case Controller Message.
			add_journal_entry(
				$journal_title_data,
				'Test Suite Error',
				"$test_name:$test_point",
				'TS_CONTROLLER_MSG',
				"$journal_file:$line_number",
				'',
				$line_text
			);
		}
		elsif ($line_code == 70) {
			# Code 70 -- Scenario Information.
			# Not interesting to us.
		}
		elsif ($line_code == 80) {
			# Code 80 -- Test Case End.
			unless ($test_name) {
				# Test case end found when no test case was started;
				# issue a warning.
				add_journal_entry(
					$journal_title_data,
					'Warning',
					'',
					'J_TC_NOT_STARTED',
					"$journal_file:$line_number",
					'',
					''
				);
				next;
			}

			# Warn if the test case seems to be empty.
			if ($test_points_in_this_case == 0) {
				add_journal_entry(
					$journal_title_data,
					'Warning',
					$test_name,
					'R_EMPTY_TEST_CASE',
					"$journal_file:$line_number",
					'',
					''
				);
			}

			# Clear the test name, number, and messages.
			$test_name = '';
			$test_number = 0;
			$test_messages = '';
			$test_points_in_this_case = 0;
		}
		elsif ($line_code == 90) {
			# Code 90 -- User Abort.
			add_journal_entry(
				$journal_title_data,
				'Test Suite Error',
				'',
				'TS_ABORTED',
				"$journal_file:$line_number",
				'',
				''
			);
		}
		elsif ($line_code == 100) {
			# Code 100 -- Captured Output.
			# Not interesting to us.
		}
		elsif ($line_code == 110) {
			# Code 110 -- Build Start.
			# Not interesting to us.
		}
		elsif ($line_code == 130) {
			# Code 130 -- Build End.
			# Not interesting to us.
		}
		elsif ($line_code == 200) {
			# Code 200 -- Test Purpose Start.
			# The structure of the line is as follows:
			# 200|activitynum pointnum hh:mm:ss|comment
			unless ($line =~ /^\d*\|(\d*)\s*(\d*)/) {
				add_journal_entry(
					$journal_title_data,
					'Warning',
					'',
					'J_SYNTAX_ERROR',
					"$journal_file:$line_number",
					'',
					$line
				);
				next;
			}

			# Check if the previous test point was properly finished.
			if ($test_point > 0) {
				if ($test_point == $2) {
					# Some test suites have a strange habit of re-starting
					# the same test point more than once.
					add_journal_entry(
						$journal_title_data,
						'Notice',
						"$test_name:$test_point",
						'J_TP_RESTARTED',
						"$journal_file:$line_number",
						'',
						''
					);
				}
				else {
					# Notify about a disappeared test.
					add_journal_entry(
						$journal_title_data,
						'Warning',
						"$test_name:$2",
						'J_TP_NOT_FINISHED',
						"$journal_file:$line_number",
						'',
						"Previous test point number: $test_point"
					);
				}
			}

			# Clear the messages.
			$test_messages = '';

			# Store the test point number.
			$test_point = $2;

			# Increment the number of test points.
			++$test_points_in_this_case;
		}
		elsif ($line_code == 220) {
			# Code 220 -- Test Purpose Result.
			# The structure of the line is as follows:
			# 220|activitynum pointnum result time|comment
			unless ($line =~ /^\d*\|(\d*)\s*(\d*)\s*(\d*)/) {
				add_journal_entry(
					$journal_title_data,
					'Warning',
					'',
					'J_SYNTAX_ERROR',
					"$journal_file:$line_number",
					'',
					$line
				);
				next;
			}

			# Print a warning if the test purpose seems to never been started.
			if ($test_point == 0) {
				add_journal_entry(
					$journal_title_data,
					'Warning',
					"$test_name:$test_point",
					'J_TP_NOT_STARTED',
					"$journal_file:$line_number",
					'',
					''
				);
			}

			# Print a warning if the test purpose number is different from
			# what we expected.
			elsif ($test_point != $2) {
				add_journal_entry(
					$journal_title_data,
					'Warning',
					"$test_name:$test_point",
					'J_TP_NOT_STARTED',
					"$journal_file:$line_number",
					'',
					"Started with TP $test_point; ended with TP $2."
				);
			}

			# Get the result code.
			my $result_code = $3;
			my $testcase_id = "$1_$2";

			# Find the meaning of the result code (severity).
			my $severity = $result_code_map{$result_code};
			if ($severity) {
				my $cat_infos = guess_problem_subcategory($test_messages, $journal_title_data->{'TEST_TOOL'});
				my $testcase_file_id = "$test_file_id.fip_$testcase_id";
				my $severity_upd = $severity;
				if ($severity eq 'FIP') {
					if ($manual_tests_data->{'TESTCASES'}->{$testcase_file_id}) {
						$severity_upd .= (($manual_tests_data->{'TESTCASES'}->{$testcase_file_id} eq 'yes') ? ': Passed' : ': Failed');
					}
					else {
						++$remaining_fips;
					}
					# Store information about FIPs for saving it into a separate file later.
					foreach my $info (@$cat_infos) {
						$fips{$info->{'SUBCATEGORY'}}->{'fip_'.$testcase_id} = ($fip_details{$info->{'SUBCATEGORY'}} ? $fip_details{$info->{'SUBCATEGORY'}}->{'DESCRIPTION'} : $info->{'MESSAGE'});
					}
				}
				foreach my $info (@$cat_infos) {
					add_journal_entry(
						$journal_title_data,
						$severity_upd,
						"$test_name:$test_point",
						$info->{'SUBCATEGORY'},
						$info->{'ELEMENT'},
						'',
						$info->{'MESSAGE'}
					);
				}
			}
			else {
				add_journal_entry(
					$journal_title_data,
					'Failed',
					"$test_name:$test_point",
					'R_UNKNOWN_RESULT',
					"$journal_file:$line_number",
					'',
					"Result code: $result_code.\nMessages from the test:\n$test_messages"
				);
			}

			# Clear the test point number.
			$test_point = 0;
			$journal_empty = 0;
		}
		elsif ($line_code == 310) {
			# 310 -- Clean Start.
			# Not interesting to us.
		}
		elsif ($line_code == 320) {
			# 320 -- Clean End.
			# Not interesting to us.
		}
		elsif ($line_code == 400) {
			# 400 -- Invocable Component Start.
			# Not interesting to us.
		}
		elsif ($line_code == 410) {
			# 410 -- Invocable Component End.
			# Not interesting to us.
		}
		elsif ($line_code == 510) {
			# 510 -- Error reported by the underlying test suite.
			add_journal_entry(
				$journal_title_data,
				'Test Suite Error',
				"$test_name:$test_point",
				'TS_ERROR',
				'',
				'',
				$line_text
			);
		}
		elsif ($line_code == 520) {
			# 520 -- Message from the underlying test suite.

			# Coalesce consequent rows of spaces in the message
			# to a single space. This improves matching with markers.
			my $message = $line_text;
			$message =~ s/\s+/ /g;

			# Add the message to the accumulated messages.
			$test_messages .= "$message\n";
		}
		elsif ($line_code == 600) {
			# 600 -- Parallel Start.
			# Not interesting to us.
		}
		elsif ($line_code == 620) {
			# 620 -- Parallel End.
			# Not interesting to us.
		}
		elsif ($line_code == 630) {

			# 630 -- Implied Sequential Start.
			# Not interesting to us.
		}
		elsif ($line_code == 640) {
			# 640 -- Implied Sequential End.
			# Not interesting to us.
		}
		elsif ($line_code == 700) {
			# 700 -- Repeat Start.
			# Not interesting to us.
		}
		elsif ($line_code == 720) {
			# 720 -- Repeat End.
			# Not interesting to us.
		}
		elsif ($line_code == 730) {
			# 730 -- Timed Loop Start.
			# Not interesting to us.
		}
		elsif ($line_code == 740) {
			# 740 -- Timed Loop End.
			# Not interesting to us.
		}
		elsif ($line_code == 750) {
			# 750 -- Random Start.
			# Not interesting to us.
		}
		elsif ($line_code == 760) {
			# 760 -- Random End.
			# Not interesting to us.
		}
		elsif ($line_code == 800) {
			# 800 -- Remote Start.
			# Not interesting to us.
		}
		elsif ($line_code == 820) {
			# 820 -- Remote End.
			# Not interesting to us.
		}
		elsif ($line_code == 830) {
			# 830 -- Distributed Start.
			# Not interesting to us.
		}
		elsif ($line_code == 840) {
			# 840 -- Distributed End.
			# Not interesting to us.
		}
		elsif ($line_code == 900) {
			# 900 -- Test Case Controller End.
			# Not interesting to us.
		}
		else {
			# Other lines are notified to the user and otherwise ignored.
			add_journal_entry(
				$journal_title_data,
				'Notice',
				"$test_name:$test_point",
				'J_FORMAT_ERROR',
				"$journal_file:$line_number",
				'',
				"Unknown code number $line_code. Journal line:\n$line"
			);
		}
	}

	if (scalar(keys(%fips)) > 0) {
		if (open(FIPS_FILE, ">$journal_file.FIPS")) {
			foreach my $cat (keys(%fips)) {
				print FIPS_FILE "[$cat]\n";
				if ($fip_details{$cat}) {
					print FIPS_FILE "CATEGORY_TITLE: $fip_details{$cat}->{'TITLE'}\n";
				}
				elsif ($subcategories{$cat}) {
					print FIPS_FILE "CATEGORY_TITLE: $subcategories{$cat}->{'CATEGORY'}\n";
				}
				foreach my $fip_id (keys %{$fips{$cat}}) {
					print FIPS_FILE <<DATA;
TESTCASE $fip_id
$fips{$cat}->{$fip_id}
TESTCASE_END
DATA
				}
				print FIPS_FILE "\n";
			}
			close(FIPS_FILE);
		}
		else {
			print "Failed to create file $journal_file.FIPS:\n$!\n";
		}
	}

	if ($journal_empty) {
		$journal_title_data->{'IS_RESULT'} = 0;
	}
	close(JOURNAL);
}

# Reads and parses the manual tests journal.
sub parse_manual_journal() {
	read_manual_tests("$result_path/$result_dir/ManualTests");
	return if (scalar(keys %spec_manual_tests) == 0);

	my $journal_file = 'test_journals/manual_tests.simplejournal';
	my $journal_title = 'Manual Tests';
	my $journal_anchor = 'manual_tests';

	add_elem_to_tree($journal_title, $journal_anchor, $journal_file);

	# Store the journal title and anchor.
	push @journal_titles, $journal_title;
	$journal_title_from_anchor{$journal_anchor} = $journal_title;

	$journal_data{$journal_title} = {
		'JOURNAL_FILE' => $journal_file,
		'TEST_TOOL' => '&lt;Manual&gt;',
		'START_TIME' => '&lt;Not applicable&gt;',
		'PROBLEMS_LIST' => [],
		'IS_RESULT' => 1,
		'STATISTICS' => {
			'PASSES' => 0,
			'FAILS' => 0,
			'WARNS' => 0,
			'FIPS' => 0,
			'NOTICES' => 0,
			'QUELLED' => 0,
			'SKIPPED' => 0
		}
	};

	# List of problems in the current journal
	my $journal_title_data = $journal_data{$journal_title};

	my $manual_journal_data = $manual_tests_data->{'TESTCASES'};
	foreach my $test_id (@spec_manual_tests_sort) {
		my $optional = $spec_manual_tests{$test_id}->{'OPTIONAL'};
		if ($manual_journal_data->{$test_id}) {
			if ($manual_journal_data->{$test_id} eq 'yes') {
				add_journal_entry(
					$journal_title_data,
					'Manual: Passed',
					$test_id,
					'R_MANUAL',
					'',
					($optional ? '<i>Optional requirement</i>' : ''),
					'{noreplace}'.$spec_manual_tests{$test_id}->{'MESSAGE'}.'{/noreplace}'
				);
			}
			else {
				add_journal_entry(
					$journal_title_data,
					($optional ? 'Warning' : 'Failed'),
					$test_id,
					'R_MANUAL',
					'',
					($optional ? '<i>Optional requirement</i>, <b>failed</b>' : ''),
					'{noreplace}'.$spec_manual_tests{$test_id}->{'MESSAGE'}.'{/noreplace}'
				);
			}
		}
		else {
			add_journal_entry(
				$journal_title_data,
				'Manual: Not performed',
				$test_id,
				'R_MANUAL',
				'',
				($optional ? '<i>Optional requirement</i>' : ''),
				'{noreplace}'.$spec_manual_tests{$test_id}->{'MESSAGE'}.'{/noreplace}'
			);
		}
	}
}

sub sum_stats() {
	my %stats = (
		'PASSES' => 0,
		'FAILS' => 0,
		'WARNS' => 0,
		'FIPS' => 0,
		'NOTICES' => 0,
		'QUELLED' => 0,
		'SKIPPED' => 0
	);
	my $total = 0;
	foreach my $journal (keys %journal_data) {
		if ($journal_data{$journal}->{'IS_RESULT'}) {
			foreach my $st (keys %stats) {
				$stats{$st} += $journal_data{$journal}->{'STATISTICS'}{$st};
				$total += $stats{$st};
			}
		}
	}
	$stats{'TOTAL'} = $total;
	return \%stats;
}

# Returns maximum value from the array.
sub max(@) {
	if (scalar(@_) == 0) {
		return undef;
	}
	my $res = $_[0];
	foreach (@_) {
		$res = $_ if ($res < $_);
	}
	return $res;
}

# Returns array with number of passes, failes and warnings in the test.
# The test is specified via unique HTML anchor.
sub report_testcases_info($) {
	my $journal_anchor = $_[0];
	if (!$journal_anchor) {
		return (-1, -1, -1, -1);
	}
	if (!$journal_title_from_anchor{$journal_anchor}) {
		return (-1, -1, -1, -1);
	}
	my $journal_data_stats = $journal_data{$journal_title_from_anchor{$journal_anchor}}->{'STATISTICS'};
	return ($journal_data_stats->{'PASSES'}, $journal_data_stats->{'FAILS'}, $journal_data_stats->{'WARNS'}, $journal_data_stats->{'FIPS'});
}

my $separator_line = '<img src="images/environment-spacer.gif" width="100%" height="1" style="background-color: white;" alt="" />';
my @verdicts       = ('&nbsp;',  '&nbsp;',  'Success', 'Warnings', 'Failed',  'FIPs');
my @verdict_colors = ('#e7e7e7', '#d0d0ff', '#40ff40', '#ffff40',  '#ff6060', '#a0a0ff');

sub report_failures_verdict($$) {
	my ($journal_anchor, $is_root) = @_;

	if (!$journal_anchor) {
		return ($is_root ? 1 : 0); # Nothing
	}

	if (!$journal_title_from_anchor{$journal_anchor}) {
		return ($is_root ? 1 : 0); # Nothing
	}

	my $journal_title = $journal_title_from_anchor{$journal_anchor};

	if (!$journal_data{$journal_title}->{'IS_RESULT'}) {
		return 4; # Failed (No result)
	}

	my $total_stats = $journal_data{$journal_title}->{'STATISTICS'};
	if ($total_stats->{'FAILS'} > 0) {
		return 4; # Failed
	}
	elsif (($total_stats->{'WARNS'} > 0) or ($total_stats->{'FIPS'} > 0)) {
		return 3; # Warning
	}
	else {
		return 2; # Success
	}
}

sub report_failures_cell($$) {
	my ($journal_anchor, $is_root) = @_;
	my $bg_color = $verdict_colors[$is_root ? 1 : 0];

	if (!$journal_anchor) {
		return "<div style=\"background-color: $bg_color;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />&nbsp;<br /></div>$separator_line";
	}

	if (!$journal_title_from_anchor{$journal_anchor}) {
		return "<div style=\"background-color: $bg_color;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />&nbsp;<br /></div>$separator_line";
	}

	my $journal_title = $journal_title_from_anchor{$journal_anchor};

	my $status = '';
	my $total_stats = $journal_data{$journal_title}->{'STATISTICS'};
	if (!$journal_data{$journal_title}->{'IS_RESULT'}) {
		$bg_color = $verdict_colors[4];
		$status = '<span title="\'No result\' means the tests failed to run, interrupted or crashed during the execution.">No Result<span class="view_srv">&nbsp; <a href="tests_help.pl#No_result" title="Help" target="_blank"><img src="themes/'.$appchecker_branch{'DIR'}.'/images/question.png" alt="[?]" width="16" height="16" /></a></span></span>';
	}
	elsif (($journal_anchor eq 'manual_tests') && ($total_stats->{'FIPS'} > 0)) {
		$bg_color = $verdict_colors[5];
		$status = 'Incomplete';
	}
	elsif ($total_stats->{'FAILS'} > 0) {
		$bg_color = $verdict_colors[4];
		$status = 'Failures: '.$total_stats->{'FAILS'};
	}
	elsif ($total_stats->{'WARNS'} > 0) {
		$bg_color = $verdict_colors[3];
		$status = 'Warnings: '.$total_stats->{'WARNS'};
	}
	else {
		$bg_color = $verdict_colors[2];
		$status = 'Success';
	}

	return "<div style=\"background-color: $bg_color;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />$status<img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" /><br /></div>$separator_line";
}

sub report_passes_cell($$) {
	my ($journal_anchor, $is_root) = @_;
	my $bg_color = $verdict_colors[$is_root ? 1 : 0];
	my ($passes, $fails, $warns, $fips) = report_testcases_info($journal_anchor);
	if ($passes == -1) {
		return "<div style=\"background-color: $bg_color;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />&nbsp;<br /></div>$separator_line";
	}
	elsif (($journal_anchor eq 'manual_tests') && ($fips > 0)) {
		return "<div style=\"background-color: $bg_color;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />&nbsp;<br /></div>$separator_line";
	}
	else {
		return "<div style=\"background-color: $bg_color;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />Passed: $passes<img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" /><br /></div>$separator_line";
	}
}

my $failures_cell = "";
my $passes_cell = "";

sub print_tree_tab($$$$);       # Prototype for recursive function
sub print_tree_tab($$$$) {
	my ($REPORT, $j_tree, $parent_id, $tab) = @_;
	my %sum_my_info = ("passes" => 0, "test_status" => 0, "passes_cell" => "", "failures_cell" => "");
	my @children = sort keys %$j_tree;
	for (my $i = 0; $i <= $#children; $i++) {
		my $elem_name = $children[$i];
		my $elem_id = esc_id($elem_name, {'NOESC' => '/'});
		$elem_id =~ s/\//:/g;
		$elem_id = "$parent_id:$elem_id";
		my $img_f_type;
		my $img_plus_minus;
		my $img_suffix;
		my $tab_new;
		if ($i < $#children) {
			$img_suffix = "d";
			$tab_new = "$tab<img src=\"images/tree_d.png\" alt=\"[ ]\" width=\"21\" height=\"21\" />";
		}
		else {
			$img_suffix = "";
			$tab_new = "$tab<img src=\"images/tree_n.png\" alt=\"[ ]\" width=\"21\" height=\"21\" />";
		}

		if ($j_tree->{$elem_name}{"type"} eq "dir") {
			$img_f_type = "<img src=\"images/tree_dir.png\" alt=\"[D]\" width=\"21\" height=\"21\" />";
			$img_plus_minus = "<img src=\"images/tree_minus_u$img_suffix.png\" alt=\"[-]\" width=\"21\" height=\"21\" id=\"icon-$elem_id\" style=\"cursor:pointer;\" onclick=\"javascript:show_hide('$elem_id', '_u$img_suffix');\" />";
		}
		elsif ($j_tree->{$elem_name}{"type"} eq "pkg") {
			$img_f_type = "<img src=\"images/tree_pkg.png\" alt=\"[P]\" width=\"21\" height=\"21\" />";
			if ($j_tree->{$elem_name}{"children"}) {
				$img_plus_minus = "<img src=\"images/tree_minus_u$img_suffix.png\" alt=\"[-]\" width=\"21\" height=\"21\" id=\"icon-$elem_id\" style=\"cursor:pointer;\" onclick=\"javascript:show_hide('$elem_id', '_u$img_suffix');\" />";
			}
			else {
				$img_plus_minus = "<img src=\"images/tree_dr$img_suffix.png\" alt=\"[ ]\" width=\"21\" height=\"21\" />";
			}
		}
		else {
			$img_f_type = "<img src=\"images/tree_file.png\" alt=\"[F]\" width=\"21\" height=\"21\" />";
			$img_plus_minus = "<img src=\"images/tree_dr$img_suffix.png\" alt=\"[ ]\" width=\"21\" height=\"21\" />";
		}

		my $elem_link = esc_html($elem_name);
		if ($j_tree->{$elem_name}{"journal_anchor"}) {
			$elem_link = "<a href=\"#".$j_tree->{$elem_name}{"journal_anchor"}."\" title=\"Click here to proceed to details\">$elem_link</a>";
			push @journal_anchors_sorted, $j_tree->{$elem_name}{"journal_anchor"};
		}
		print $REPORT "<div style=\"background-color: #e7e7e7; white-space: nowrap;\">$tab$img_plus_minus$img_f_type<b>$elem_link</b><img src=\"images/environment-spacer.gif\" width=\"5\" height=\"21\" alt=\"\" /><br /></div>$separator_line";

		# Calculate passes and failures for the element and its children
		my $local_passes_cell;
		my $local_failures_cell;
		my ($local_test_passes, $local_test_failures, $local_test_warnings, $local_test_fips) = report_testcases_info($j_tree->{$elem_name}{"journal_anchor"});
		my $local_test_status;
		if ($local_test_passes == -1) {
			$local_passes_cell = "<div style=\"background-color: #e7e7e7;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />&nbsp;<br /></div>$separator_line";
			$local_failures_cell = "<div style=\"background-color: #e7e7e7;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />&nbsp;<br /></div>$separator_line";
			$local_test_passes = 0;
			$local_test_status = 0;
		}
		else {
			$local_passes_cell = report_passes_cell($j_tree->{$elem_name}{"journal_anchor"}, 0);
			$local_test_status = report_failures_verdict($j_tree->{$elem_name}{"journal_anchor"}, 0);
			$local_failures_cell = report_failures_cell($j_tree->{$elem_name}{"journal_anchor"}, 0);
			$sum_my_info{"passes"} += $local_test_passes;
			$sum_my_info{"test_status"} = max($local_test_status, $sum_my_info{"test_status"});
		}
		if ($j_tree->{$elem_name}{"children"}) {
			print $REPORT "<div id=\"root-$elem_id\">\n";
			my %sum_children_info = print_tree_tab($REPORT, $j_tree->{$elem_name}{"children"}, $elem_id, $tab_new);
			print $REPORT "</div>\n";
			$local_test_passes += $sum_children_info{"passes"};
			$sum_my_info{"passes"} += $local_test_passes;
			$local_passes_cell = "<div style=\"display: none;\" id=\"pass-total-$elem_id\">\n<div style=\"background-color: #e7e7e7;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />Passed: ".$local_test_passes."<img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" /><br /></div>$separator_line</div>\n<div id=\"pass-children-$elem_id\">\n".$local_passes_cell.$sum_children_info{"passes_cell"}."</div>\n";
			$sum_my_info{"passes_cell"} .= $local_passes_cell;

			$local_test_status = max($local_test_status, $sum_children_info{"test_status"});
			$sum_my_info{"test_status"} = max($local_test_status, $sum_my_info{"test_status"});
			$local_failures_cell = "<div style=\"display: none;\" id=\"fail-total-$elem_id\">\n<div style=\"background-color: ".$verdict_colors[$local_test_status].";\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />".$verdicts[$local_test_status]."<img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" /><br /></div>$separator_line</div>\n<div id=\"fail-children-$elem_id\">\n".$local_failures_cell.$sum_children_info{"failures_cell"}."</div>\n";
			$sum_my_info{"failures_cell"} .= $local_failures_cell;
		}
		else {
			$sum_my_info{"passes_cell"} .= $local_passes_cell;
			$sum_my_info{"failures_cell"} .= $local_failures_cell;
		}
	}
	return %sum_my_info;
}

sub print_tree($$) {
	my ($REPORT, $j_tree) = @_;
	print $REPORT "<table border=\"0\" cellpadding=\"0\" cellspacing=\"0\"><tr><td><img src=\"images/environment-spacer.gif\" width=\"2\" height=\"1\" alt=\"\" /></td><td valign=\"top\">\n";
	foreach my $elem_name (sort keys %$j_tree) {
		my $elem_id = $elem_name;
		$elem_id =~ s/^PKG: /PKG\//;
		$elem_id = esc_id($elem_id, {'NOESC' => '/'});
		$elem_id =~ s/\//:/g;
		my $img_f_type;
		my $img_plus_minus;
		if ($j_tree->{$elem_name}{"type"} eq "dir") {
			$img_f_type = "<img src=\"images/tree_dir.png\" alt=\"[D]\" width=\"21\" height=\"21\" />";
			$img_plus_minus = "<img src=\"images/tree_plus.png\" alt=\"[+]\" width=\"21\" height=\"21\" id=\"icon-$elem_id\" style=\"cursor:pointer;\" onclick=\"javascript:show_hide('$elem_id', '');\" />";
		}
		elsif ($j_tree->{$elem_name}{"type"} eq "pkg") {
			$img_f_type = "<img src=\"images/tree_pkg.png\" alt=\"[P]\" width=\"21\" height=\"21\" />";
			if ($j_tree->{$elem_name}{"children"}) {
				$img_plus_minus = "<img src=\"images/tree_plus.png\" alt=\"[+]\" width=\"21\" height=\"21\" id=\"icon-$elem_id\" style=\"cursor:pointer;\" onclick=\"javascript:show_hide('$elem_id', '');\" />";
			}
			else {
				$img_plus_minus = "<img src=\"images/tree_n.png\" alt=\"[ ]\" width=\"21\" height=\"21\" />";
			}
		}
		else {
			$img_f_type = "<img src=\"images/tree_file.png\" alt=\"[F]\" width=\"21\" height=\"21\" />";
			$img_plus_minus = "<img src=\"images/tree_n.png\" alt=\"[ ]\" width=\"21\" height=\"21\" />";
		}
		my $elem_link = esc_html($elem_name);
		if ($j_tree->{$elem_name}{"journal_anchor"}) {
			$elem_link = "<a href=\"#".$j_tree->{$elem_name}{"journal_anchor"}."\" title=\"Click here to proceed to details\">$elem_link</a>";
			push @journal_anchors_sorted, $j_tree->{$elem_name}{"journal_anchor"};
		}
		print $REPORT "<div style=\"background-color: #d0d0ff; white-space: nowrap;\">\n$img_plus_minus$img_f_type<b>$elem_link</b><img src=\"images/environment-spacer.gif\" width=\"5\" height=\"21\" alt=\"\" /><br />\n</div>$separator_line";

		# Calculate passes and failures for the element and its children
		my $local_passes_cell;
		my $local_failures_cell;
		my ($local_test_passes, $local_test_failures, $local_test_warnings, $local_test_fips) = report_testcases_info($j_tree->{$elem_name}{"journal_anchor"});
		my $local_test_status;
		if ($local_test_passes == -1) {
			$local_passes_cell = "<div style=\"background-color: #d0d0ff;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />&nbsp;<br /></div>$separator_line";
			$local_failures_cell = "<div style=\"background-color: #d0d0ff;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />&nbsp;<br /></div>$separator_line";
			$local_test_passes = 0;
			$local_test_status = 1;
		}
		else {
			$local_passes_cell = report_passes_cell($j_tree->{$elem_name}{"journal_anchor"}, 1);
			$local_test_status = report_failures_verdict($j_tree->{$elem_name}{"journal_anchor"}, 1);
			$local_failures_cell = report_failures_cell($j_tree->{$elem_name}{"journal_anchor"}, 1);
		}
		if ($j_tree->{$elem_name}{"children"}) {
			print $REPORT "<div id=\"root-$elem_id\" style=\"display: none;\">\n";
			my %sum_children_info = print_tree_tab($REPORT, $j_tree->{$elem_name}{"children"}, $elem_id, "<img src=\"images/tree_n.png\" alt=\"[ ]\" width=\"21\" height=\"21\" />");
			print $REPORT "</div>\n";
			$local_test_passes += $sum_children_info{"passes"};
			$local_passes_cell = "<div id=\"pass-total-$elem_id\">\n<div style=\"background-color: #d0d0ff;\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />Passed: ".$local_test_passes."<img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" /><br /></div>$separator_line</div>\n<div style=\"display: none;\" id=\"pass-children-$elem_id\">\n".$local_passes_cell.$sum_children_info{"passes_cell"}."</div>\n";
			$local_test_status = max($local_test_status, $sum_children_info{"test_status"});
			$local_failures_cell = "<div id=\"fail-total-$elem_id\">\n<div style=\"background-color: ".$verdict_colors[$local_test_status].";\"><img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" />".$verdicts[$local_test_status]."<img src=\"images/environment-spacer.gif\" alt=\"\" width=\"5\" height=\"21\" /><br /></div>$separator_line</div>\n<div style=\"display: none;\" id=\"fail-children-$elem_id\">\n".$local_failures_cell.$sum_children_info{"failures_cell"}."</div>\n";
		}
		$passes_cell .= $local_passes_cell;
		$failures_cell .= $local_failures_cell;
	}
	print $REPORT <<DATA;
</td>
<td><img src="images/environment-spacer.gif" width="1" height="1" alt="" /></td>
<td valign="top" style="white-space: nowrap;">
$failures_cell
</td>
<td><img src="images/environment-spacer.gif" width="1" height="1" alt="" /></td>
<td valign="top" style="white-space: nowrap;">
$passes_cell
</td>
</tr></table>
DATA
}

sub host_info() {
	my $path_url = esc_url("$result_path/$result_dir", {'NOESC' => '/'});
	return <<DATA;
<h3>Host Info</h3>

<table class="wikitable">
  <tr>
    <th style="text-align: left;">Host Name</th>
    <td>$host_name</td>
  </tr>
  <tr>
    <th style="text-align: left;">Host OS</th>
    <td>$host_os</td>
  </tr>
  <tr>
    <th style="text-align: left;">Host Machine</th>
    <td>$host_machine</td>
  </tr>
  <tr>
    <th style="text-align: left;">Test Architecture</th>
    <td>$test_arch</td>
  </tr>
  <tr>
    <th style="text-align: left;">Host Kernel</th>
    <td>$host_kernel</td>
  </tr>
</table>
The complete console output log is available <a href="/get.pl$path_url/test_logs/tests_output" class="view_srv" target="_blank">here</a><a href="test_logs/tests_output" class="view_as_is" target="_blank">here</a>.
<br />
<br />

DATA
}

# Creates the HTML report
sub compile_report() {
	my $REPORT;
	if (!open($REPORT, ">$result_path/$result_dir/test_log.html")) {
		print "Couldn't open file '$result_path/$result_dir/test_log.html' for writing: $!";
		return;
	}

	my $bugreport_email = 'lsb-appcheck-support@linuxfoundation.org';

	my $auto_test_problems = 0;
	my $manual_tests_needed = 0;
	foreach (keys %journal_data) {
		my $stats = $journal_data{$_}->{'STATISTICS'};
		if ($_ ne 'Manual Tests') {
			if ($stats->{'FAILS'} or $stats->{'WARNS'} or $stats->{'FIPS'}) {
				$auto_test_problems = 1;
			}
		}
		if ($stats->{'FIPS'}) {
			$manual_tests_needed = 1;
		}
	}

	my $run_manual_tests = '';
	if ($manual_tests_needed) {
		my $test_run_url = esc_url($result_dir);
		$run_manual_tests = <<DATA;
<div class="view_srv" style="background-color: #ff4040; border: solid 2px darkred; font-weight: bold; width: 20em; color: white; margin-top: 1.5em; margin-bottom: 0.5em; padding: 3px; text-align: center;">
<a href="tests_manual.pl?test_run=$test_run_url" style="color: white; text-decoration: underline;">RUN MANUAL TESTS</a>
</div>
<p class="view_srv">There are manual checks that have not been performed yet! For getting the application certification, it necessary to complete them too.</p>
DATA
	}

	my $cat_group_selected = ($auto_test_problems ? ' selected="selected"' : '');
	my $comp_group_selected = ($auto_test_problems ? '' : ' selected="selected"');
	my $cat_group_visibility = ($auto_test_problems ? '' : 'display: none;');
	my $comp_group_visibility = ($auto_test_problems ? 'display: none;' : '');

	# Print Title
	my $application_name_html = esc_html($application_name);
	my $title_highlight = (($appchecker_branch{'STD'} eq 'LSB') ? 'blue' : '#c00000');
	print $REPORT <<DATA;
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en" dir="ltr">
<head>
  <title>$appchecker_branch{'STD'} $spec_version Test Results for $application_name_html on $test_arch</title>
  <meta http-equiv="Content-Type" content="text/html; charset=utf-8" />
  <style type="text/css">
    .view_srv {
        display: none;
    }
    table {
      empty-cells: show;
    }
    table.wikitable, table.wikitable-sep {
      margin: 1em 1em 1em 0;
      background: #f9f9f9;
    }
    table.wikitable {
      border: 1px #aaaaaa solid;
      border-collapse: collapse;
    }
    table.wikitable th, table.wikitable td {
      border: 1px #aaaaaa solid;
      padding: 0.2em;
      padding-left: 0.4em;
      padding-right: 0.4em;
    }
    table.wikitable-sep th, table.wikitable-sep td {
      border-bottom: 1px #aaaaaa solid;
      border-right: 1px #aaaaaa solid;
      padding: 0.2em;
      padding-left: 0.4em;
      padding-right: 0.4em;
    }
    table.wikitable-sep th.btop, table.wikitable-sep td.btop {
      border-top: 1px #aaaaaa solid;
    }
    table.wikitable-sep th.bleft, table.wikitable-sep td.bleft {
      border-left: 1px #aaaaaa solid;
    }
    table.wikitable th.left {
      background: #f2f2f2;
      text-align: left;
      padding-left: 0.4em;
      padding-right: 0.4em;
    }
    table.wikitable th, table.wikitable-sep th {
      background: #f2f2f2;
      text-align: center;
      padding-left: 0.4em;
      padding-right: 0.4em;
    }
    .title-highlight {
      color: $title_highlight;
    }
  </style>
</head>

<body>

<script language="javascript" type="text/javascript">
// <![CDATA[
// Preload some images that are not initially visible
var preload_image_names = new Array('tree_d.png', 'tree_dir.png', 'tree_dr.png', 'tree_drd.png', 'tree_file.png', 'tree_minus.png', 'tree_minus_u.png', 'tree_minus_ud.png', 'tree_n.png', 'tree_pkg.png', 'tree_plus.png', 'tree_plus_u.png', 'tree_plus_ud.png', 'report_plus.png', 'report_minus.png');
var preload_images_arr = new Array();
for (var i=0; i<preload_image_names.length; ++i) {
	preload_images_arr[i] = new Image;
	preload_images_arr[i].src = 'images/' + preload_image_names[i];
}

function show_hide(id, icon_suffix) {
	var elem = document.getElementById('root-' + id);
	var icon = document.getElementById('icon-' + id);
	var passTotal = document.getElementById('pass-total-' + id);
	var passChildren = document.getElementById('pass-children-' + id);
	var failTotal = document.getElementById('fail-total-' + id);
	var failChildren = document.getElementById('fail-children-' + id);
	if (elem && icon && passTotal && passChildren && failTotal && failChildren) {
		var to_show = (elem.style.display == 'none');
		if (to_show) {
			elem.style.display = '';
			passTotal.style.display = 'none';
			passChildren.style.display = '';
			failTotal.style.display = 'none';
			failChildren.style.display = '';
			icon.setAttribute('src', 'images/tree_minus' + icon_suffix + '.png');
			icon.setAttribute('alt', '[-]');
		}
		else {
			elem.style.display = 'none';
			passTotal.style.display = '';
			passChildren.style.display = 'none';
			failTotal.style.display = '';
			failChildren.style.display = 'none';
			icon.setAttribute('src', 'images/tree_plus' + icon_suffix + '.png');
			icon.setAttribute('alt', '[+]');
		}
	}
}

function bug_report(journal_path, test_app, test_case) {
	if (confirm('A new E-mail to $bugreport_email will now be created.\\nPlease specify all the necessary details and attach the journal files:\\n\\n' + journal_path + '\\n\\nContinue?')) {
		document.location = 'mailto:$bugreport_email?Subject=AppChecker:%20Problem%20Report%20(' + test_case + ')&Body=File tested: ' + test_app + '%0A%0A';
	}
}

function submit_problem_info(journal_path, test_app, test_case) {
	if (confirm('A new E-mail to $bugreport_email will now be created.\\nPlease specify all the necessary details and attach the journal files:\\n\\n' + journal_path + '\\n\\nContinue?')) {
		document.location = 'mailto:$bugreport_email?Subject=AppChecker:%20Submit%20a%20Comment%20(' + test_case + ')&Body=File tested: ' + test_app + '%0A%0AComments:%0A';
	}
}

function show_hide_msg(img, id) {
	var elem = document.getElementById(id);
	if (!elem)
		return;
	var collapsed = (elem.style.display == 'none');
	if (collapsed) {
		elem.style.display = '';
		img.setAttribute('src', 'images/report_minus.png');
	}
	else {
		elem.style.display = 'none';
		img.setAttribute('src', 'images/report_plus.png');
	}
}

function switch_group(name) {
	var elem_cat = document.getElementById('by-category');
	var elem_comp = document.getElementById('by-component');
	if (elem_cat && elem_comp) {
		if (name == 'component') {
			elem_cat.style.display = 'none';
			elem_comp.style.display = '';
		}
		else {
			elem_cat.style.display = '';
			elem_comp.style.display = 'none';
		}
	}
}
// ]]>
</script>

<h1 class="view_as_is">$appchecker_branch{'STD'} $spec_version Test Results for <span class="title-highlight">$application_name_html</span> on <span class="title-highlight">$test_arch</span></h1>

<!-- Test Report Header -->
$run_manual_tests
<br />

<table class="bartable">
  <tr>
    <td align="right"><b>Group Test Results by:</b></td>
    <td align="right">
      <select name="std_filter_libs" id="std_filter_libs" onchange="javascript:switch_group(this.value);" style="width: 12.5em;">
        <option value="category"$cat_group_selected>Problem Category</option>
        <option value="component"$comp_group_selected>Component</option>
      </select>
    </td>
  </tr>
</table>
<br />

<div id="by-component" style="$comp_group_visibility">
<h3>Test Summary</h3>
DATA

	# Print the index tree
	print_tree($REPORT, \%journal_tree);
	print $REPORT "<br />\n\n";
	print $REPORT host_info();

	print $REPORT <<DATA;
<h1>Detailed Reports by Components</h1>
DATA

	# Test journals
	my $first_journal = 1;
	foreach my $journal_anchor (@journal_anchors_sorted) {
		my $journal_title;
		if ($journal_title_from_anchor{$journal_anchor}) {
			$journal_title = $journal_title_from_anchor{$journal_anchor};
		}
		else {
			next;
		}

		# User-friendly file path
		my $journal_title_desc = $journal_title;
		$journal_title_desc =~ s/^\|//;
		$journal_title_desc =~ s/\|$//;
		$journal_title_desc =~ s/\|/ => /g;
		$journal_title_desc = esc_html($journal_title_desc);
		if ($first_journal) {
			$first_journal = 0;
		}
		else {
			print $REPORT "<hr />\n<br />\n";
		}
		print $REPORT "<h2><a name=\"$journal_anchor\"></a><font color=\"#336699\">$journal_title_desc</font></h2>\n";

		print $REPORT generate_journal_report($journal_anchor, $journal_title, $journal_title_desc);
	}
	print $REPORT <<DATA;
</div>

<div id="by-category" style="$cat_group_visibility">
<h3>Test Summary</h3>
DATA
	# Print the category list
	my @filtered_list = ();
	print $REPORT <<DATA;
<table cellpadding="1" cellspacing="1" border="0">
DATA
	if (!$auto_test_problems) {
		unshift @categories_sort, 'Automatic Tests';
		$cat_problems{'Automatic Tests'} = {
			'PROBLEMS' => [],
			'FAILS' => 0,
			'WARNS' => 0,
			'FIPS' => 0
		};
	}
	foreach my $cat (@categories_sort) {
		next if (!defined($cat_problems{$cat}) or (($cat ne 'Automatic Tests') and (scalar(@{$cat_problems{$cat}->{'PROBLEMS'}}) == 0)));
		push @filtered_list, $cat;
		my $lc_cat = lc($cat);
		$lc_cat =~ s/[^a-z0-9\-_:.]/-/g;
		my $cell_txt;
		my $color;
		if (($cat eq 'Manual Tests') and $cat_problems{$cat}->{'FIPS'}) {
			$cell_txt = 'Incomplete';
			$color = $verdict_colors[5];
		}
		elsif ($cat_problems{$cat}->{'FAILS'}) {
			$cell_txt = 'Failures: '.$cat_problems{$cat}->{'FAILS'};
			$color = $verdict_colors[4];
		}
		elsif ($cat_problems{$cat}->{'WARNS'}) {
			$cell_txt = 'Warnings: '.$cat_problems{$cat}->{'WARNS'};
			$color = $verdict_colors[3];
		}
		elsif ($cat_problems{$cat}->{'FIPS'}) {
			$cell_txt = 'FIPs: '.$cat_problems{$cat}->{'FIPS'};
			$color = $verdict_colors[5];
		}
		else {
			$cell_txt = 'No problems';
			$color = $verdict_colors[0];
		}
		print $REPORT <<DATA;
  <tr>
    <td style="background-color: $verdict_colors[1];"><img src="images/tree_file.png" alt="" width="21" height="21" /><b><a href="#$lc_cat" title="Click here to proceed to details">$cat</a></b></td>
    <td style="background-color: $color;">$cell_txt</td>
  </tr>
DATA
	}
	print $REPORT <<DATA;
</table>
<br />
DATA
	print $REPORT host_info();

	print $REPORT <<DATA;
<h1>Detailed Reports by Problem Category</h1>
DATA

	# Print the problems
	foreach my $cat (@filtered_list) {
		my $lc_cat = lc($cat);
		$lc_cat =~ s/[^a-z0-9\-_:.]/-/g;
		my $elem_col_name = $categories{$cat}->{'ELEM_COL_NAME'};
		print $REPORT <<DATA;
<h2><a name="$lc_cat"></a><font color="#336699">$cat</font></h2>
$categories{$cat}->{'DESCRIPTION'}
DATA
		next if (scalar(@{$cat_problems{$cat}->{'PROBLEMS'}}) == 0);
		print $REPORT <<DATA;
<table class="wikitable-sep" cellpadding="0" cellspacing="0">
  <tr>
    <th class="btop bleft" style="width: 17px;"></th>
    <th class="btop" style="width: 25em;">Component</th>
    <th class="btop" style="width: 5em;">Type</th>
    <th class="btop" style="width: 10em;">$elem_col_name</th>
    <th class="btop" style="width: 25em;">Comments</th>
  </tr>
DATA
		# Sort problems list by element
		my @problems_list_sorted = sort {
			$journal_data{$a->{'JOURNAL_TITLE'}}->{'PROBLEMS_LIST'}->[$a->{'PROBLEM_IDX'}]->{'ELEMENT'}
			cmp
			$journal_data{$b->{'JOURNAL_TITLE'}}->{'PROBLEMS_LIST'}->[$b->{'PROBLEM_IDX'}]->{'ELEMENT'}
		} @{$cat_problems{$cat}->{'PROBLEMS'}};
		foreach (my $i = 0; $i < scalar(@problems_list_sorted); ++$i) {
			my $journal_title = $problems_list_sorted[$i]->{'JOURNAL_TITLE'};
			my $journal_title_desc = esc_js(esc_html($journal_title));
			my $id = "$lc_cat-$i";
			my $idx = $problems_list_sorted[$i]->{'PROBLEM_IDX'};
			my $problem = $journal_data{$journal_title}->{'PROBLEMS_LIST'}->[$idx];
			my $severity = $problem->{'SEVERITY'};
			my $severity_color = '';
			my $subcategory = $subcategories{$problem->{'SUBCATEGORY'}};
			my $td_id = ' id="'.esc_id($problem->{'TEST_CASE'}).'"';
			if (($severity =~ m/Error/i) or
				($severity =~ m/Failed/i) or
				($severity =~ m/Unresolved/i) or
				($severity =~ m/Uninitiated/i) or
				($severity =~ m/Unreported/i) or
				($severity =~ m/Time Expired/i)) {
				$severity_color = " background-color: $verdict_colors[4];";
			}
			elsif ($severity =~ m/Warning/i) {
				$severity_color = " background-color: $verdict_colors[3];";
			}
			elsif ($severity =~ m/^FIP$/i) {
				$severity_color = " background-color: $verdict_colors[5];";
				$severity = "<span title=\"'FIP' stands for 'Further Information Provided'.\">$severity<span class=\"view_srv\">&nbsp; <a href=\"tests_help.pl#FIP\" title=\"Help\" target=\"_blank\"><img src=\"themes/$appchecker_branch{'DIR'}/images/question.png\" alt=\"[?]\" width=\"16\" height=\"16\" /></a></span></span>";
			}
			elsif ($severity =~ m/^Not performed$/i) {
				$severity_color = " background-color: $verdict_colors[5];";
			}
			elsif ($severity =~ m/Passed/i) {
				$severity_color = " background-color: #40ff40;";
			}
			my $element = $problem->{'ELEMENT'};
			if ($subcategory->{'ELEM_LINK'} and ($element ne '') and ($element ne '...')) {
				my $link_element = esc_url($element);
				$element = "<a href=\"$subcategory->{'ELEM_LINK'}$link_element\" target=\"_blank\">$element</a>";
			}
			my $comment = ($problem->{'COMMENT'} ? '<td>'.$problem->{'COMMENT'}.'</td>' : '<td style="color: #a0a0a0; white-space: nowrap;">No information</td>');
			my $test_msg = html_excl($problem->{'DETAILS'});
			my $component = $journal_title;
			$component =~ s/^.*\.rpm\|([^|]+)/RPM:$1/;
			$component =~ s/^.*\.deb\|([^|]+)/DEB:$1/;
			$component =~ s/^.*\.(tar)\|([^|]+)/TAR:$2/;
			$component =~ s/^.*\.(tar\.gz|tgz)\|([^|]+)/TGZ:$2/;
			$component =~ s/^.*\.(tar\.bz2|tbz)\|([^|]+)/TBZ:$2/;
			$component =~ s/^PKG:.*\|([^|]+)/PKG:$1/;
			$component =~ s/\|//g;
			$component = esc_html($component);
			$journal_title =~ s/\|$//;
			$journal_title =~ s/\|/ => /g;
			$journal_title = esc_html($journal_title);
			my $test_case = esc_js(esc_html($problem->{'TEST_CASE'}));
			my $archive_path = esc_js(esc_html("$result_path/$result_dir/$result_archive"));
			print $REPORT <<DATA;
  <tr>
    <td class="bleft" rowspan="2" style="text-align: center; vertical-align: top;"><img src="images/report_plus.png" alt="[+]" width="17" height="17" style="cursor: pointer;" onclick="javascript:show_hide_msg(this, '$id');" /></td>
    <td style="white-space: nowrap;" title="$journal_title">$component</td>
    <td style="text-align: center; white-space: nowrap;$severity_color"$td_id><b>$severity</b></td>
    <td style="white-space: nowrap;">$element</td>
    $comment
  </tr>
  <tr>
    <td colspan="4" id="$id" style="display: none;">
      $subcategory->{'MESSAGE'}<br/><div style='font-family:monospace; padding-left: 3em;'>$test_msg</div><div align="right"><input type="button" value="Submit Your Comment" style="width: 12em;" onclick="javascript:submit_problem_info('$archive_path', '$journal_title_desc', '$test_case');" title="Press this button to submit additional information about this problem." />&nbsp;<input type="button" value="Report Bug in the Test" style="width: 12em;" onclick="javascript:bug_report('$archive_path', '$journal_title_desc', '$test_case');" title="Press this button if you beleive the problem is in the test, not in your system." /></div>
    </td>
  </tr>
DATA
		}
		print $REPORT "</table>\n";
	}

	print $REPORT <<DATA;
</div>
</body>
</html>
DATA
	close($REPORT);
}

sub generate_journal_report($$$) {
	my ($journal_anchor, $journal_title, $journal_title_desc) = @_;
	my $journal_title_data = $journal_data{$journal_title};

	if (!defined($journal_title_data->{'IS_RESULT'})) {
		return <<DATA;
<h3 style="color: #a00000;">No result.</h3>
Journal file for <b>$journal_title_desc</b> could not be opened. This usually means that the corresponding test failed to run, interrupted or crashed during the execution.
DATA
	}

	my $stat_data = $journal_title_data->{'STATISTICS'};
	my $problems_list = $journal_title_data->{'PROBLEMS_LIST'};
	my $journal_path_href = esc_url("$result_path/$result_dir", {'NOESC' => '/'});
	my $journal_file_href = esc_url($journal_title_data->{'JOURNAL_FILE'}, {'NOESC' => '/'});
	my $verdict_idx = report_failures_verdict($journal_anchor, 0);
	my $fail_color = ($stat_data->{'FAILS'} ? ' background-color: '.$verdict_colors[4] : '');
	my $warn_color = ($stat_data->{'WARNS'} ? ' background-color: '.$verdict_colors[3] : '');
	my $fip_color = ($stat_data->{'FIPS'} ? ' background-color: '.$verdict_colors[3] : '');
	my $res = '';
	$res = <<DATA;
<h3>Test Statictics</h3>

<table cellpadding="0" cellspacing="0" border="0">
  <tr>
    <td style="vertical-align: top;">
      <table class="wikitable">
        <tr>
          <th style="text-align: left; width: 7em;">Passed tests</th>
          <td style="text-align: right; width: 4em;">$stat_data->{'PASSES'}</td>
        </tr>
        <tr>
          <th style="text-align: left;">Failed tests</th>
          <td style="text-align: right;$fail_color">$stat_data->{'FAILS'}</td>
        </tr>
DATA
	if ($stat_data->{'WARNS'}) {
		$res .= <<DATA;
        <tr>
          <th style="text-align: left;">Warnings</th>
          <td style="text-align: right;$warn_color">$stat_data->{'WARNS'}</td>
        </tr>
DATA
	}
	if ($stat_data->{'FIPS'}) {
		$res .= <<DATA;
        <tr>
          <th style="text-align: left;">FIPs</th>
          <td style="text-align: right;$fip_color">$stat_data->{'FIPS'}</td>
        </tr>
DATA
	}
	$res .= <<DATA;
        <tr>
          <th style="text-align: left;">Verdict</th>
          <td style="text-align: right; background-color: $verdict_colors[$verdict_idx];">$verdicts[$verdict_idx]</td>
        </tr>
      </table>
    </td>
    <td width="20"></td>
    <td style="vertical-align: top;">
      <table class="wikitable">
        <tr>
          <th style="text-align: left;">Test Tool</th>
          <td style="text-align: right;">$journal_title_data->{'TEST_TOOL'} (<a href="/get.pl$journal_path_href/$journal_file_href" class="view_srv" target="_blank">open journal file</a><a href="$journal_file_href" class="view_as_is" target="_blank">open journal file</a>)</td>
        </tr>
        <tr>
          <th style="text-align: left;">Test Start Time</th>
          <td style="text-align: right;">$journal_title_data->{'START_TIME'}</td>
        </tr>
      </table>
    </td>
  </tr>
</table>

DATA
	if (!$journal_title_data->{'IS_RESULT'}) {
		$res .= <<DATA;
<h3 style="color: #a00000;">No result.</h3>
Journal file for <b>$journal_title_desc</b> is empty or contains no test cases. This usually means that the corresponding test failed to run, interrupted or crashed during the execution.
DATA
	}
	elsif (scalar(@$problems_list) == 0) {
		$res .= "<h3>No problems detected</h3>\n";
	}
	else {
		$res .= <<DATA;
<h3>List of Problems</h3>

<table class="wikitable-sep" cellpadding="0" cellspacing="0">
  <tr>
    <th class="btop bleft" style="width: 17px;"></th>
    <th class="btop" style="width: 16em;">Category</th>
    <th class="btop" style="width: 5em;">Type</th>
    <th class="btop" style="width: 10em;">Element</th>
    <th class="btop" style="width: 25em;">Comments</th>
  </tr>
DATA
		$journal_title_desc = esc_js($journal_title_desc);
		foreach (my $i = 0; $i < scalar(@$problems_list); ++$i) {
			my $id = "$journal_anchor-$i";
			my $severity = $problems_list->[$i]->{'SEVERITY'};
			my $severity_color = '';
			my $subcategory = $subcategories{$problems_list->[$i]->{'SUBCATEGORY'}};
			my $is_fail = 0;
			my $is_warn = 0;
			my $is_fip = 0;
			my $td_id = ' id="'.esc_id($problems_list->[$i]->{'TEST_CASE'}).'"';
			if (($severity =~ m/Error/i) or
				($severity =~ m/Failed/i) or
				($severity =~ m/Unresolved/i) or
				($severity =~ m/Uninitiated/i) or
				($severity =~ m/Unreported/i) or
				($severity =~ m/Time Expired/i)) {
				$severity_color = " background-color: $verdict_colors[4];";
				$is_fail = 1;
			}
			elsif ($severity =~ m/Warning/i) {
				$severity_color = " background-color: $verdict_colors[3];";
				$is_warn = 1;
			}
			elsif ($severity =~ m/^FIP$/i) {
				$severity_color = " background-color: $verdict_colors[5];";
				$severity = "<span title=\"'FIP' stands for 'Further Information Provided'.\">$severity<span class=\"view_srv\">&nbsp; <a href=\"tests_help.pl#FIP\" title=\"Help\" target=\"_blank\"><img src=\"themes/$appchecker_branch{'DIR'}/images/question.png\" alt=\"[?]\" width=\"16\" height=\"16\" /></a></span></span>";
				$is_fip = 1;
			}
			elsif ($severity =~ m/^Not performed$/i) {
				$severity_color = " background-color: $verdict_colors[5];";
				$is_fip = 1;
			}
			elsif ($severity =~ m/Passed/i) {
				$severity_color = " background-color: #40ff40;";
			}
			my $element = $problems_list->[$i]->{'ELEMENT'};
			if ($subcategory->{'ELEM_LINK'} and ($element ne '') and ($element ne '...')) {
				$element = "<a href=\"$subcategory->{'ELEM_LINK'}$element\" target=\"_blank\">$element</a>";
			}
			my $comment = ($problems_list->[$i]->{'COMMENT'} ? '<td>'.$problems_list->[$i]->{'COMMENT'}.'</td>' : '<td style="color: #a0a0a0; white-space: nowrap;">No information</td>');
			my $category = $subcategory->{'CATEGORY'};
			my $test_msg = html_excl($problems_list->[$i]->{'DETAILS'});
			my $test_case = esc_js(esc_html($problems_list->[$i]->{'TEST_CASE'}));
			my $archive_path = esc_js(esc_html("$result_path/$result_dir/$result_archive"));
			$res .= <<DATA;
  <tr>
    <td class="bleft" rowspan="2" style="text-align: center; vertical-align: top;"><img src="images/report_plus.png" alt="[+]" width="17" height="17" style="cursor: pointer;" onclick="javascript:show_hide_msg(this, '$id');" /></td>
    <td style="white-space: nowrap;">$category</td>
    <td style="text-align: center; white-space: nowrap;$severity_color"$td_id><b>$severity</b></td>
    <td style="white-space: nowrap;">$element</td>
    $comment
  </tr>
  <tr>
    <td colspan="4" id="$id" style="display: none;">
      $subcategory->{'MESSAGE'}<br/><div style='font-family:monospace; padding-left: 3em;'>$test_msg</div><div align="right"><input type="button" value="Submit Your Comment" style="width: 12em;" onclick="javascript:submit_problem_info('$archive_path', '$journal_title_desc', '$test_case');" title="Press this button to submit additional information about this problem." />&nbsp;<input type="button" value="Report Bug in the Test" style="width: 12em;" onclick="javascript:bug_report('$archive_path', '$journal_title_desc', '$test_case');" title="Press this button if you beleive the problem is in the test, not in your system." /></div>
    </td>
  </tr>
DATA
			# Add problem into problems-in-category list
			if (!defined($cat_problems{$category})) {
				$cat_problems{$category}->{'FAILS'} = 0;
				$cat_problems{$category}->{'WARNS'} = 0;
				$cat_problems{$category}->{'FIPS'} = 0;
				$cat_problems{$category}->{'PROBLEMS'} = [];
			}
			push @{$cat_problems{$category}->{'PROBLEMS'}}, {
				'JOURNAL_TITLE' => $journal_title,
				'PROBLEM_IDX' => $i
			};
			$cat_problems{$category}->{'FAILS'} += $is_fail;
			$cat_problems{$category}->{'WARNS'} += $is_warn;
			$cat_problems{$category}->{'FIPS'} += $is_fip;
		}
		$res .= "</table>\n";
	}
	return $res;
}

################################################################################
# Start execution

parse_cmd_line();
read_host_info();
generate_report();
